# auto scripts
def nb_to_py(
    notebookp: str,
    test: bool=False,
    validate: bool=True,
    sep_step: str='## step',
    notebook_suffix: str='_v',
    ):
    """notebook to script.

    Args:
        notebookp (str): path to the notebook.
        sep_step (str, optional): separator marking the start of a step. Defaults to "## step".
        notebook_suffix (str, optional): suffix of the notebook file to be considered as a "task".
        test (bool, optional): test mode. Defaults to False.
        validate (bool, optional): validate. Defaults to True.

    TODOs: 
        1. Add `check_outputs` parameter to only filter out non-executable code (i.e. tests) if False else edit the code.
    """

    from roux.lib.set import unique
    import pandas as pd
    print(notebookp)
    d1=read_dict(notebookp,fmt='json',encoding="utf8")
    if len(d1['cells'])==0:
        return
    funs=[]
    fun=''
    for di,d in enumerate(d1['cells']):
        if d['cell_type']=='markdown' and len(d['source'])!=0: 
            if d['source'][0].startswith(sep_step):
                funs.append(fun)
                fun=''
        if d['cell_type']=='code': 
#             print(d['source'])
            fun+=('\n'.join(d['source']) if len(d['source'])>1 else f"\n{d['source'][0]}\n" if len(d['source'])==1 else '')
#             print(fun)
    if di==len(d1['cells'])-1:
        funs.append(fun)
    funs=[s.split('## trash')[0] for s in funs if '\nto_' in s]
    
    if len(funs)==0:
        logging.error(f'{basename(notebookp)}: no functions found')
        if not test:
            return ''
        else:
            return '',pd.DataFrame()
    df1=pd.DataFrame(pd.Series(funs,name='code raw'))
    if validate:#check_outputs:
        def get_path_output(s):
            s='\n'.join([line for line in s.split('\n') if not line.startswith('#')])
            for f in ['to_table','to_dict']:
                if f in s and not 'to_dict()' in s:
                    return s.split(f)[1].split(',')[1].split(')')[0].replace("'",'')
    
        df1.loc[:,'path output']=df1['code raw'].apply(get_path_output)
        df1=df1.log.dropna(subset=['path output'])
        if test:
            from roux.lib.dfs import to_table
            to_table(df1,'test/notebook2packagescript.tsv')
        df1.loc[:,'parameter output']=df1['path output'].apply(lambda x: basenamenoext(x)+'p')
        def get_paths_input(s):
            s='\n'.join([line for line in s.split('\n') if not line.startswith('#')])
            paths=[]
            for f in ['read_table(','read_dict(']:
                if f in s:
                    paths.append(s.split(f)[1].split(',')[0].split(')')[0].replace("'",''))
            return paths
        df1.loc[:,'paths input']=df1['code raw'].apply(get_paths_input)
        df1.loc[:,'parameters input']=df1['paths input'].apply(lambda x: [basenamenoext(s)+'p' for s in x])
        df1.loc[:,'parameters']=df1.apply(lambda x: ['cfg']+x['parameters input']+[x['parameter output']],axis=1)
        if any(df1['parameters'].apply(lambda x: len(unique(x))!=len(x))):
            logging.error(f'{notebookp}: duplicate parametter/s')
        if df1['path output'].apply(lambda x: basename(dirname(x))).unique().shape[0]!=1:
            logging.error(f"{notebookp}: should be a single output directory. {','.join(df1['path output'].apply(lambda x: basename(dirname(x))).unique().tolist())}")
        else:
            if df1['path output'].apply(lambda x: basename(dirname(x))).unique()[0].replace('data','')!=basename(notebookp).split(notebook_suffix)[0]:
                logging.error(f"{notebookp}: output directory should match notebook directory. {df1['path output'].apply(lambda x: basename(dirname(x))).unique()[0].replace('data','')}!={basename(notebookp).split(notebook_suffix)[0]}")

        df1.loc[:,'function name']=df1.apply(lambda x: f"get{x.name:02d}_{x['parameter output']}",axis=1)
        df1.loc[:,'function line']=df1.apply(lambda x: f"def {x['function name']}({','.join(x['parameters'])}):",axis=1)

        def get_code(x):
            from roux.lib.str import replacemany
            code=replacemany(x['code raw'],{f"'{x['path output']}'":x['parameter output'],
            f"\"{x['path output']}\"":x['parameter output']})
            code=replacemany(code,dict(zip([f"'{s}'" for s in x['paths input']],x['parameters input'])))
            code=x['function line']+'\n'+'    '+code.replace('\n','\n    ')
            return code.replace('\n    \n    ','\n    ')
        df1.loc[:,'code']=df1.apply(get_code,axis=1)

    code='from roux.global_imports import *\n'+'\n\n'.join(df1['code'].tolist())
    if not test:
        return code
    else:
        return code,df1

import re
def sort_stepns(
    l: list
    ) -> list:
    """Sort steps (functions) of a task (script).

    Args:
        l (list): list of steps.

    Returns:
        list: sorted list of steps.
    """
    l=[s for s in l if bool(re.search('\d\d_',s))]    
    l_=[int(re.findall('\d\d',s)[0]) for s in l if bool(re.search('\d\d_',s))]
    return sort_list_by_list(l,l_)



import argh
from roux.lib.io import backup,version,to_zip

## begin
import sys
parser = argh.ArghParser()
parser.add_commands([backup,version,to_zip])

if __name__ == '__main__':
    parser.dispatch()


"""
io_df -> io_dfs -> io_files
"""
# paths
from roux.lib.dfs import *
from roux.lib.dict import read_dict,to_dict # overwrite to_dict from df
from roux.lib.sys import * #is_interactive_notebook,basenamenoext,makedirs,get_all_subpaths
from roux.lib.str import replacemany
from shutil import copyfile
import logging

## paths
def read_ps(ps,test=True):
    """Read a list of paths.
    
    Parameters:
        ps (list|str): list of paths or a string with wildcard/s.
        test (bool): testing.

    Returns:
        ps (list): list of paths.
    """
    if isinstance(ps,str): 
        if '*' in ps:
            ps=glob(ps)
        else:
            ps=[ps]
    ps=sorted(ps)
    if test:
        ds1=pd.Series({p:p2time(p) if exists(p) else np.nan for p in ps}).sort_values().dropna()
        if len(ds1)>1:
            from roux.lib.str import get_suffix
            d0=ds1.iloc[[0,-1]].to_dict()
            for k_,k,v  in zip(['oldest','latest'],get_suffix(*d0.keys(),common=False),d0.values()):
                logging.info(f"{k_}: {k}\t{v}")
        elif len(ds1)==0:
            logging.warning('paths do not exist.')
    return ps

def to_path(
    s,
    replacewith='_',
    verbose=False,
    coff_len_escape_replacement=100,
    ):
    """Normalise a string to be used as a path of file.
    
    Parameters:
        s (string): input string.
        replacewith (str): replace the whitespaces or incompatible characters with.
        
    Returns:
        s (string): output string.
    """
    import re
    s=re.sub(r'(/)\1+',r'\1',s) # remove multiple /'s
    if max([len(s_) for s_ in s.split('/')])<coff_len_escape_replacement:
        s=(re.sub(r'[^\w+/.+-=]',replacewith, s)
           .replace('+',replacewith) 
           .strip(replacewith)
           )
        s=re.sub(r'(_)\1+',r'\1',s) # remove multiple _'s
    else:
        if verbose:
            logging.info("replacements not done; possible long IDs in the path.")
    return s.replace(f'/My{replacewith}Drive/','/My Drive/') # google drive
#     return re.sub('\W+',replacewith, s.lower() )

# alias to be deprecated in the future
make_pathable_string=to_path
# get_path=to_path

def to_outp(ps,outd=None,outp=None,suffix=''):
    """Infer output path based on the list of paths.
    
    Parameters:
        ps (list): list of paths.
        outd (str): path of the output directory.
        outp (str): path of the output file.
        suffix (str): suffix of the filename.
    
    Returns:
        outp (str): path of the output file. 
    """
    if not outp is None:
        return outp
    from roux.lib.str import get_prefix
    # makedirs(outd)
    ps=read_ps(ps)
    pre=get_prefix(ps[0],ps[-1], common=True)
    if not outd is None:
        pre=outd+(basename(pre) if basename(pre)!='' else basename(dirname(pre)))
    outp=f"{pre}_{suffix}{splitext(ps[0])[1]}"
    return outp

def get_encoding(p):
    """Get encoding of a file.
    
    Parameters:
        p (str): file path
        
    Returns:
        s (string): encoding.
    """
    import chardet
    with open(p, 'rb') as f:
        result = chardet.detect(f.read())
    return result['encoding']                

import shutil
def to_zip(p, outp=None,fmt='zip'):
    """Compress a file/directory.
    
    Parameters:
        p (str): path to the file/directory.
        outp (str): path to the output compressed file.
        fmt (str): format of the compressed file.
    
    Returns:
        outp (str): path of the compressed file.
    """
    def _to_zip(source, destination=None,
          fmt='zip'):
        """
        Zip a folder.
        Ref:
        https://stackoverflow.com/a/50381250/3521099
        """
        if destination is None:
            destination=source.rsplit('/')+"."+fmt
        from os import sep
        base = basename(destination)
        fmt = base.split('.')[-1]
        name = base.replace('.'+fmt,'')
    #     print(base,name,fmt)
        archive_from = dirname(source)
        if archive_from=='':
            archive_from='./'
    #     archive_to = basename(source.strip(sep))
    #     print(archive_from,archive_to)
        shutil.make_archive(name, fmt, archive_from, 
    #                         archive_to
                           )
        shutil.move(f'{name}.{fmt}', destination)
        return destination
    
    if isinstance(p,str):
        if isdir(p):
            return _to_zip(p, destination=outp, fmt=fmt)        
    ps=read_ps(p)
    import tempfile
    with tempfile.TemporaryDirectory() as outd:
        _=[copyfile(p,f"{outd}/{basename(p)}") for p in ps]
        return _to_zip(outd+'/', destination=outp, fmt=fmt)
    
def read_zip(p,file_open=None, fun_read=None):
    """Read the contents of a zip file.
    
    Parameters:
        p (str): path of the file.
        file_open (str): path of file within the zip file to open.
        fun_read (object): function to read the file.

    """
    from io import BytesIO
    from zipfile import ZipFile
    from urllib.request import urlopen
    if p.startswith('http') or p.startswith('ftp'):
        p=BytesIO(urlopen(p).read())
    file = ZipFile(p)
    if file_open is None:
        return file
    else:
        if fun_read is None:
            return file.open(file_open)
        else:
            return fun_read(file.open(file_open))
        
def get_version(suffix=''):
    """Get the time-based version string.
    
    Parameters:
        suffix (string): suffix.
    
    Returns:
        version (string): version.
    """
    return 'v'+get_time().replace('_','')+'_'+suffix

def version(p,outd=None,
            **kws):
    """Get the version of the file/directory.
    
    Parameters:
        p (str): path.
        outd (str): output directory.
        
    Keyword parameters:
        kws (dict): provided to `get_version`.
    
    Returns: 
        version (string): version.        
    """
    p=p.rstrip("/")
    if outd is None:
        outd=f"{dirname(p)}{'/' if dirname(p)!='' else ''}"
    if isdir(p):
        outp=f"{outd}.{get_version(basename(p),**kws)}"
    else:
        outp=f"{outd}.{get_version(basenamenoext(p),**kws)}{splitext(p)[1]}"    
    logging.info(p,outp)
    shutil.move(p,outp)
    return outp

def backup(p,outd,
           versioned=False,
           suffix='',
           zipped=False,
           move_only=False,
           test=True,
          no_test=False
          ):
    """Backup a directory
    
    Steps:
        0. create version dir in outd
        1. move ps to version (time) dir with common parents till the level of the version dir
        2. zip or not
        
    Parameters:
        p (str): input path.
        outd (str): output directory path.
        versioned (bool): custom version for the backup (False).
        suffix (str): custom suffix for the backup ('').
        zipped (bool): whether to zip the backup (False).
        test (bool): testing (True).
        no_test (bool): no testing (False).
                
    TODO:
        1. Chain to if exists and force.
        2. Option to remove dirs
            find and move/zip
            "find -regex .*/_.*"
            "find -regex .*/test.*"
    """
    info(p,outd)
    if no_test:
        test=False
    logging.warning(f"test={test}")
    from roux.lib.set import unique
    ps=read_ps(p)
    assert(len(ps)!=0)
#     if test:print(ps)
    ps=[ p+'/' if isdir(p) else p for p in ps]
#     if test:print(ps)
    from roux.lib.sys import get_time
#     if timed:
    outd2=outd+'/.'+get_version(suffix)#'/_v'+get_time()+'_'+(suffix+'_' if not suffix is None else '')
    # create directoried in outd
    outds=unique([dirname(dirname(p)) if isdir(p) else dirname(p) for p in ps])
#     if test:print(outds)
    outds=[replacemany(p,{outd:outd2})+'/' if not move_only else outd2+p for p in outds]
#     if test:print(outds)
    l1=[(p,replacemany(p,{outd:outd2})) if not move_only else (p,outd2+p) for p in ps]
#     if test:print(l1)
    l1=[(p1,dirname(dirname(p2)) if p2.endswith('/') else p2) for p1,p2 in l1]    
    if test:
        return l1
    assert(len(outds)!=0)
    _=[makedirs(p) for p in outds]
    l2=[shutil.move(*t) for t in l1]
    # zip
    if zipped:
        to_zip(outd2, f"{outd2}.zip")
        return f"{outd2}.zip"
    else:
        return l2

def read_url(url):
    """Read text from an URL.
    
    Parameters:
        url (str): URL link.
        
    Returns:
        s (string): text content of the URL.
    """
    from urllib.request import urlopen
    f = urlopen(url)
    myfile = f.read()
    return str(myfile)

def download(url,path=None,outd='data/database',
             force=False,
             verbose=True):
    """Download a file.
    
    Parameters:
        url (str): URL. 
        path (str): custom output path (None)
        outd (str): output directory ('data/database').
        force (bool): overwrite output (False).
        verbose (bool): verbose (True).
        
    Returns: 
        path (str): output path (None)
    """
    def get_download_date(path):
        import os
        import datetime
        t = os.path.getctime(path)
        return str(datetime.datetime.fromtimestamp(t))
    if path is None:
        path=replacemany(url,
               {'https://':'',
                'http://':'',
               })
        path=f"{outd}/{path}"
    if not exists(path) or force:
        import urllib.request
        makedirs(path,exist_ok=True)
        urllib.request.urlretrieve(url, path)
    elif verbose:
        logging.info(f"downloaded on: {get_download_date(path)}")
    return path

## dfs
def post_read_table(df1,clean,tables,
                    verbose=True,
                    **kws_clean):
    """Post-reading a table.
    
    Parameters: 
        df1 (DataFrame): input dataframe.
        clean (bool): whether to apply `clean` function. 
        tables ()
        verbose (bool): verbose.
    
    Keyword parameters:
        kws_clean (dict): paramters provided to the `clean` function. 
        
    Returns:
        df (DataFrame): output dataframe. 
    """
    if clean:
        df1=df1.rd.clean(**kws_clean)
    if tables==1 and verbose:
        df1=df1.log()
    return df1
    
from roux.lib.text import get_header
def read_table(p,
               ext=None,
               clean=True,
               filterby_time=None,
               check_paths=True, # read files in the path column
               test=False,
               params={},
               kws_clean={},
               kws_cloud={},
               tables=1,
               verbose=True,
               **kws_read_tables
              ):
    """
    Table/s reader.
    
    Parameters:
        p (str): path of the file. It could be an input for `read_ps`, which would include strings with wildcards, list etc. 
        ext (str): extension of the file (default: None meaning infered from the path).
        clean=(default:True).
        filterby_time=None).
        check_paths (bool): read files in the path column (default:True). 
        test (bool): testing (default:False).
        params: parameters provided to the 'pd.read_csv' (default:{}). For example
            params['columns']: columns to read.
        kws_clean: parameters provided to 'rd.clean' (default:{}).
        kws_cloud: parameters for reading files from google-drive (default:{}).
        tables: how many tables to be read (default:1).
        verbose: verbose (default:True).
                
    Keyword parameters:
        kws_read_tables (dict): parameters provided to `read_tables` function. For example:
            replaces_index (object|dict|list|str): for example, 'basenamenoext' if path to basename.
            drop_index (bool): whether to drop the index column e.g. `path` (default: True).
            colindex (str): the name of the column containing the paths (default: 'path')
    
    Returns:
        df (DataFrame): output dataframe. 
        
    Examples:
        1. For reading specific columns only set `params=dict(columns=list)`.

        2. Reading a vcf file.
            p='*.vcf|vcf.gz'
            read_table(p,
                       params_read_csv=dict(
                       #compression='gzip',
                       sep='\t',comment='#',header=None,
                       names=replacemany(get_header(path,comment='#',lineno=-1),['#','\n'],'').split('\t'))
                       )
    """
    if isinstance(p,list) or (isinstance(p,str) and ('*' in p)):
        if (isinstance(p,str) and ('*' in p)):
            ps=read_ps(p,test=False)
            if exists(p.replace('/*','')):
                logging.warning(f"exists: {p.replace('/*','')}")
        elif isinstance(p,list):
            ps=p
        return read_tables(ps,params=params,
                               filterby_time=filterby_time,
                               tables=len(ps),
                               verbose=verbose,
                               **kws_read_tables)
    elif isinstance(p,str):
        ## read paths
        if check_paths and isdir(splitext(p)[0]):
            # if len(read_ps(f"{splitext(p)[0]}/*{splitext(p)[1]}",test=False))>0:
            df_=read_table(p,check_paths=False)
            if df_.columns.tolist()[-1]=='path':
                logging.info(f"paths read {len(df_['path'].tolist())}paths from the file")
                return read_table(df_['path'].tolist())
            else:
                return df_
        elif p.startswith("https://docs.google.com/file/"):
            if not 'outd' in kws_cloud:
                logging.warning("outd not found in kws_cloud")
            from roux.lib.google import download_file
            return read_table(download_file(p,**kws_cloud))
    else:
        raise ValueError(p)
    assert exists(p), f"not found: {p}"
    if len(params.keys())!=0 and not 'columns' in params:
        return post_read_table(pd.read_csv(p,**params),clean=clean,tables=tables,verbose=verbose,**kws_clean)
    else:
        if len(params.keys())==0:
            params={}
        if ext is None:
            ext=basename(p).split('.',1)[1]
        if any([s==ext for s in ['pqt','parquet']]):#p.endswith('.pqt') or p.endswith('.parquet'):
            return post_read_table(pd.read_parquet(p,engine='fastparquet',**params),
                                   clean=clean,tables=tables,verbose=verbose,**kws_clean)        
        params['compression']='gzip' if ext.endswith('.gz') else 'zip' if ext.endswith('.zip') else None
        
        if not params['compression'] is None:
            ext=ext.split('.',1)[0]
            
        if any([s==ext for s in ['tsv','tab','txt']]):
            params['sep']='\t'
        elif any([s==ext for s in ['csv']]):
            params['sep']=','            
        elif ext=='vcf':
            from roux.lib.str import replacemany
            params.update(dict(sep='\t',
                               comment='#',
                               header=None,
                               names=replacemany(get_header(path=p,comment='#',lineno=-1),['#','\n'],'').split('\t'),
                              ))
        elif ext=='gpad':
            params.update(dict(
                  sep='\t',
                  names=['DB','DB Object ID','Qualifier','GO ID','DB:Reference(s) (|DB:Reference)','Evidence Code','With (or) From',
                         'Interacting taxon ID','Date','Assigned by','Annotation Extension','Annotation Properties'],
                 comment='!',
                 ))
        else: 
            logging.error(f'unknown extension {ext} in {p}')
        if test: print(params)
        return post_read_table(pd.read_table(p,**params,),clean=clean,tables=tables,verbose=verbose,**kws_clean)            

def get_logp(ps):
    """Infer the path of the log file.
    
    Parameters:
        ps (list): list of paths.     

    Returns:
        p (str): path of the output file.     
    """
    from roux.lib.str import get_prefix
    p=get_prefix(min(ps),max(ps),common=True,clean=True)
    if not isdir(p):
        p=dirname(p)
    return f"{p}.log"

def apply_on_paths(ps,func,
                   replaces_outp=None,
                   # path=None,
                   replaces_index=None,
                   drop_index=True, # keep path
                   colindex='path',
                   filter_rows=None,
                   fast=False, 
                   progress_bar=True,
                   params={},
#                    log=True,
                   dbug=False,
                   test1=False,
                   verbose=True,
                   kws_read_table={},
                   **kws,
                  ):
    """Apply a function on list of files.
    
    Parameters:
        ps (str|list): paths or string to infer paths using `read_ps`.
        func (function): function to be applied on each of the paths.
        replaces_outp (dict|function): infer the output path (`outp`) by replacing substrings in the input paths (`p`).
        filter_rows (dict): filter the rows based on dict, using `rd.filter_rows`.
        fast (bool): parallel processing (default:False). 
        progress_bar (bool): show progress bar(default:True).
        params (dict): parameters provided to the `pd.read_csv` function.
        dbug (bool): debug mode on (default:False).
        test1 (bool): test on one path (default:False).
        kws_read_table (dict): parameters provided to the `read_table` function (default:{}).
        replaces_index (object|dict|list|str): for example, 'basenamenoext' if path to basename.
        drop_index (bool): whether to drop the index column e.g. `path` (default: True).
        colindex (str): the name of the column containing the paths (default: 'path')
    
    Keyword parameters:
        kws (dict): parameters provided to the function.

    Example:
            1. Function: 
                def apply_(p,outd='data/data_analysed',force=False):
                    outp=f"{outd}/{basenamenoext(p)}.pqt'
                    if exists(outp) and not force:
                        return
                    df01=read_table(p)
                apply_on_paths(
                ps=glob("data/data_analysed/*"),
                func=apply_,
                outd="data/data_analysed/",
                force=True,
                fast=False,
                read_path=True,
                )
    """
    def read_table_(df,read_path=False,
                    save_table=False,
                    filter_rows=None,
                    replaces_outp=None,
                    params={},
                    dbug=False,
                    verbose=True,
                    **kws_read_table,
                   ):
        p=df.iloc[0,:]['path']
        if read_path:
            if save_table:
                outp=replacemany(p, replaces=replaces_outp, replacewith='', ignore=False)
                if dbug: ic(outp)
#                 if exists(outp):
#                     if 'force' in kws:
#                         if kws['force']:
#                             return None,None
#                 else:
                return p,outp
            else:
                return p,
        else:
            df=read_table(p,params=params,verbose=verbose,**kws_read_table)
            if not filter_rows is None:
                df=df.rd.filter_rows(filter_rows)            
            return df,
    import inspect
    read_path=inspect.getfullargspec(func).args[0]=='p'
    save_table=(not replaces_outp is None) and ('outp' in inspect.getfullargspec(func).args)
    if not replaces_index is None: drop_index=False
    ps=read_ps(ps,test=verbose)
    if len(ps)==0:
        logging.error('no paths found')
        return
    if test1:
        ps=ps[:1]
        logging.warning(f"test1=True, {ps[0]}")
    if (not replaces_outp is None) and ('force' in kws):
        if not kws['force']:
            # p2outp
            p2outp={p:replacemany(p, replaces=replaces_outp, replacewith='', ignore=False) for p in ps}
            if dbug: print(p2outp)
            d_={}
            d_['from']=len(ps)
            ps=[p for p in p2outp if (not exists(p2outp[p])) or isdir(p2outp[p])]
            d_['  to']=len(ps)
            if d_['from']!=d_['  to']:
                logging.info(f"force=False, so paths reduced from: {d_['from']}")
                logging.info(f"                                to: {d_['  to']}")
    if dbug: info(ps)
    df1=pd.DataFrame({'path':ps})
    if len(df1)==0:
        logging.info('no paths remained to be processed.')
        return df1
    if fast and not progress_bar: progress_bar=True
    df2=getattr(df1.groupby('path',as_index=True),
                            f"{'parallel' if fast else 'progress'}_apply" if progress_bar else "apply"
               )(lambda df: func(*(read_table_(df,read_path=read_path,
                                                 save_table=save_table,
                                                 replaces_outp=replaces_outp,
                                                 filter_rows=filter_rows,
                                                 params=params,
                                                 dbug=dbug,
                                               verbose=verbose,
                                              **kws_read_table,)),
                                 **kws))
    if isinstance(df2,pd.Series):
        return df2
    if save_table: 
        if len(df2)!=0 and not test1:
            # save log file
            from roux.lib.set import to_list
            logp=get_logp(df2.tolist())
            to_list(df2.tolist(),logp)
            logging.info(logp)
        return df2
    # if not path is None:
    #     drop_index=False
    #     colindex,replaces_index=path
    if drop_index:
        df2=df2.rd.clean().reset_index(drop=drop_index).rd.clean()
    else:
        df2=df2.reset_index(drop=drop_index).rd.clean()
        if colindex!='path':
            df2=df2.rename(columns={'path':colindex},errors='raise')
    if not replaces_index is None:
        if isinstance(replaces_index,str):
            if replaces_index=='basenamenoext':
                replaces_index=basenamenoext
        df2[colindex]=df2[colindex].apply(lambda x: replacemany(x, replaces=replaces_index, replacewith='', ignore=False))
    return df2

def read_tables(ps,
                    fast=False,
                    filterby_time=None,
                    drop_index=True,
                    to_dict=False,
                    params={},
                    tables=None,
                    **kws_apply_on_paths,
                   ):
    """Read multiple tables.
    
    Parameters:
        ps (list): list of paths.
        fast (bool): parallel processing (default:False)
        filterby_time (str): filter by time (default:None)
        drop_index (bool): drop index (default:True)
        to_dict (bool): output dictionary (default:False)
        params (dict): parameters provided to the `pd.read_csv` function (default:{})
        tables: number of tables (default:None).
        
    Keyword parameters:
        kws_apply_on_paths (dict): parameters provided to `apply_on_paths`.
        
    Returns:
        df (DataFrame): output dataframe. 
    
    TODOs:
        Parameter to report the creation dates of the newest and the oldest files.
    """
    
    if not filterby_time is None:
        from roux.lib.sys import ps2time
        df_=ps2time(ps)
        ps=df_.loc[df_['time'].str.contains(filterby_time),'p'].unique().tolist()
        drop_index=False # see which files are read
    if not to_dict:
        df2=apply_on_paths(ps,func=lambda df: df,
                           fast=fast,
                           drop_index=drop_index,
                           params=params,
                           kws_read_table=dict(tables=tables), 
#                            kws_read_table=dict(verb=False if len(ps)>5 else True),
                           **kws_apply_on_paths)
        return df2
    else:
        return {p:read_table(p,
                             params=params) for p in read_ps(ps)}

## save table
def to_table(
    df,
    p,
    colgroupby=None,
    test=False,
    **kws
    ):
    """Save table.
    
    Parameters:
        df (DataFrame): the input dataframe. 
        p (str): output path.
        colgroupby (str|list): columns to groupby with to save the subsets of the data as separate files.
        test (bool): testing on (default:False).
        
    Keyword parameters:
        kws (dict): parameters provided to the `to_manytables` function.
    
    Returns:
        p (str): path of the output.    
    """
    if is_interactive_notebook(): test=True
    p=to_path(p)
    if df is None:
        df=pd.DataFrame()
        logging.warning(f"empty dataframe saved: {p}")
#     if len(basename(p))>100:
#         p=f"{dirname(p)}/{basename(p)[:95]}_{basename(p)[-4:]}"
#         logging.warning(f"p shortened to {p}")
    if not df.index.name is None:
        df=df.reset_index()
    if not exists(dirname(p)) and dirname(p)!='':
        makedirs(p,exist_ok=True)
    if not colgroupby is None:
        to_manytables(df,p,colgroupby,**kws)
    elif p.endswith('.tsv') or p.endswith('.tab'):
        df.to_csv(p,sep='\t',**kws)
    elif p.endswith('.pqt'):
        to_table_pqt(df,p,**kws)
    else: 
        logging.error(f'unknown extension {p}')
    return p

def to_manytables(df,p,colgroupby,
                  fmt='',
                  ignore=False,
                  **kws_get_chunks):
    """
    Save many table.
    
    Parameters:
        df (DataFrame): the input dataframe. 
        p (str): output path.
        colgroupby (str|list): columns to groupby with to save the subsets of the data as separate files.
        fmt (str): if '=' column names in the folder name e.g. col1=True.
        ignore (bool): ignore the warnings (default:False).
        
    Keyword parameters:
        kws_get_chunks (dict): parameters provided to the `get_chunks` function.
    
    Returns:
        p (str): path of the output.   
    
    TODOs:
        1. Change in default parameter: `fmt='='`.
    """
    outd,ext=splitext(p)
    if isinstance(colgroupby,str):
        colgroupby=[colgroupby]
    if colgroupby=='chunk':
        if not ignore:
            if exists(outd):
                logging.error(f"can not overwrite existing chunks: {outd}/")
            assert not exists(outd), outd
        df[colgroupby]=get_chunks(df1=df,value='right',
                                  **kws_get_chunks)
#     print(df.groupby(colgroupby).progress_apply(lambda x: f"{outd}/{x.name if not isinstance(x.name, tuple) else '/'.join(x.name)}{ext}"))
    
    if (df.loc[:,colgroupby].dtypes=='float').any():        
        logging.error('columns can not be float')
        info(df.loc[:,colgroupby].dtypes)
        return 
    elif (df.loc[:,colgroupby].dtypes=='bool').any():
        fmt='='
        logging.warning('bool column detected, fmt changed to =')
    def to_outp(names,outd,colgroupby,fmt):
        if isinstance(names, str):
            names=[names]
        d1=dict(zip(colgroupby,names))
        s1='/'.join([(f"{k}{fmt}" if fmt!='' else fmt)+f"{str(v)}" for k,v in d1.items()])
        return to_path(f"{outd}/{s1}{ext}")                
    df2=df.groupby(colgroupby).progress_apply(lambda x: to_table(x,to_outp(x.name,outd,colgroupby,fmt))).to_frame('path').reset_index()
    to_table(df2,p)
    
def to_table_pqt(df,p,engine='fastparquet',compression='gzip',**kws_pqt):
    if len(df.index.names)>1:
        df=df.reset_index()    
    if not exists(dirname(p)) and dirname(p)!='':
        makedirs(p,exist_ok=True)
    df.to_parquet(p,engine=engine,compression=compression,**kws_pqt)
    
def tsv2pqt(p):
    """Convert tab-separated file to Apache parquet. 
    
    Parameters:
        p (str): path of the input.
        
    Returns: 
        p (str): path of the output.
    """
    to_table_pqt(pd.read_csv(p,sep='\t',low_memory=False),f"{p}.pqt")
def pqt2tsv(p):
    """Convert Apache parquet file to tab-separated. 
    
    Parameters:
        p (str): path of the input.
        
    Returns: 
        p (str): path of the output.
    """    
    to_table(read_table(p),f"{p}.tsv")
    
def read_excel(p,sheet_name=None,to_dict=False,kws_cloud={},**kws):
    """Read excel file
    
    Parameters:
        p (str): path of the file. 
        sheet_name (str|None): read 1st sheet if None (default:None)
        to_dict (bool): return `dict` (default:False)
        kws_cloud (dict): parameters provided to read the file from the google drive (default:{})
    
    Keyword parameters:
        kws: parameters provided to the excel reader.
    """
#     if not 'xlrd' in sys.modules:
#         logging.error('need xlrd to work with excel; pip install xlrd')
    if p.startswith("https://docs.google.com/spreadsheets/"):
        if not 'outd' in kws_cloud:
            raise ValueError("outd not found in kws_cloud")
        from roux.lib.google import download_file
        return read_excel(download_file(p,**kws_cloud),**kws)
    if not to_dict:
        if sheet_name is None:
            xl = pd.ExcelFile(p)
#             xl.sheet_names  # see all sheet names
            if sheet_name is None:
                sheet_name=input(', '.join(xl.sheet_names))
            return xl.parse(sheet_name) 
        else:
            return pd.read_excel(p, sheet_name, **kws)
    else:
        xl = pd.ExcelFile(p)
        # see all sheet names
        sheetname2df={}
        for sheet_name in xl.sheet_names:
            sheetname2df[sheet_name]=xl.parse(sheet_name) 
        return sheetname2df
        
def to_excel(sheetname2df,outp,append=False,**kws):
    """Save excel file.
    
    Parameters:
        sheetname2df (dict): dictionary mapping the sheetname to the dataframe.
        outp (str): output path. 
        append (bool): append the dataframes (default:False).

    Keyword parameters: 
        kws: parameters provided to the excel writer.
    """
#     if not 'xlrd' in sys.modules:
#         logging.error('need xlrd to work with excel; pip install xlrd')
    makedirs(outp)
    outp=to_path(outp)
    writer = pd.ExcelWriter(outp)
    startrow=0
    for sn in sheetname2df:
        if not append:
            sheetname2df[sn].to_excel(writer,sn,index=False,**kws)
        else:
            sheetname2df[sn].to_excel(writer,startrow=startrow,index=False,**kws)  
            startrow+=len(sheetname2df[sn])+2
    writer.save()
    return outp

def to_excel_commented(
    p: str,
    d1: dict,
    outp: str=None,
    author: str='Author'
    ):
    """Add comments to the columns of excel file and save.

    Args:
        p (str): input path of excel file.
        d1 (dict): map between column names and comment e.g. description of the column.
        outp (str): output path of excel file. Defaults to None.
        author (str): author of the comments. Defaults to 'Author'.
        
    TODOs:
        1. Increase the limit on comments can be added to number of columns. Currently it is 26 i.e. upto Z1.
    """
    if outp is None:
        outp=p
        logging.warning('overwritting the input file')
    from openpyxl import load_workbook
    from openpyxl.comments import Comment
    from string import ascii_uppercase
    wb = load_workbook(filename = outp)
    for sh in wb:
        for k in [s+'1' for s in list(ascii_uppercase)+['A'+s_ for s_ in ascii_uppercase]]:
            if (not sh[k].value is None):
                if (sh[k].value in d1):
                    sh[k].comment = Comment(d1[sh[k].value],author=author)
                else:
                    logging.warning(f"no comment for column: '{sh[k].value}'")
            else:
                break
    wb.save(outp)
    wb.close()
    
## to table: validate
def check_chunks(outd,col,plot=True):
    """Create chunks of the tables.
    
    Parameters:
        outd (str): output directory.
        col (str): the column with values that are used for getting the chunks.
        plot (bool): plot the chunk sizes (default:True).
    
    Returns:
        df3 (DataFrame): output dataframe.
    """
    df1=pd.concat({p:read_table(f'{p}/*.pqt',params=dict(columns=[col])) for p in glob(outd)})
    df2=df1.reset_index(0).log.dropna()
    df3=df2.groupby('level_0')[col].nunique()
    df3.index=[s.replace(outd,'').replace('/','') for s in df3.index]
    logging.info(df3)
    if plot: 
        import seaborn as sns
        sns.swarmplot(df3)
    return df3



import numpy as np

def confidence_interval_95(x: np.array) -> float:
    """95% confidence interval.

    Args:
        x (np.array): input vector.

    Returns:
        float: output.
    """
    return 1.96*np.std(x)/np.sqrt(len(x))

def get_ci(rs,ci_type,outstr=False):
    if ci_type.lower()=='max':
        ci=max([abs(r-np.mean(rs)) for r in rs])
    elif ci_type.lower()=='sd':
        ci=np.std(rs)
    elif ci_type.lower()=='ci':
        ci=confidence_interval_95(rs)
    else:
        raise ValueError("ci_type invalid")
        return
    if not outstr:
        return ci
    else:
        return "$\pm${ci:.2f}{ci_type if ci_type!='max' else ''}"


from roux.lib.df import *
import logging

def get_cols_x_for_comparison(
    df1: pd.DataFrame,
    cols_y: list,
    cols_index: list,
    cols_drop=[],
    cols_dropby_patterns=[],
    ) -> dict:
    """
    Identify X columns.
    
    """
    ## drop columns
    df1=(df1
    .drop(cols_drop,axis=1)
    .rd.dropby_patterns(cols_dropby_patterns)
    .log.dropna(how='all',axis=1)
    )

    ## drop single value columns
    drop_cols=df1.rd.check_nunique().loc[lambda x: x==1].index.tolist()#+df1.filter(like='coexp').columns.tolist()
    df1=df1.drop(drop_cols,axis=1)
    
    ## make the dictionary with column names
    d0=dict(cols_y={
            'cont':df1.loc[:,cols_y].select_dtypes((int,float)).columns.tolist(),
            },
            cols_index=cols_index,
        )
    d0['cols_y']['desc']=list(set(cols_y) - set(d0['cols_y']['cont']))
    d0['cols_x']={}
    
    ## get continuous cols_x
    from roux.lib.stat.classify import drop_low_complexity
    df_=drop_low_complexity(df1=df1,
                        min_nunique=5,
                        max_inflation=50,
                        cols=df1.select_dtypes((int,float)).columns.tolist(),
                        cols_keep=d0['cols_y']['cont'],
                       )
    d0['cols_x']['cont']=df_.drop(d0['cols_y']['cont'],axis=1).select_dtypes((int,float)).columns.tolist()
    
    ## get non-colinear ones 
    from roux.lib.stat.corr import check_collinearity
    check_collinearity(
        df1=df1.loc[:,d0['cols_x']['cont']],
        threshold=0.7,
        colvalue='$r_s$',
        cols_variable=['variable1','variable2'],
        coff_pval=0.05,)
    
    ## get descrete x columns
    ds_=df1.rd.check_nunique().sort_values()
    l1=ds_.loc[lambda x: (x==2)].index.tolist()
    logging.info(l1)
    ds_=df1.select_dtypes((int,float)).nunique().sort_values()
    l2=ds_.loc[lambda x: (x==2)].index.tolist()
    logging.info(l2)
    d0['cols_x']['desc']=sorted(list(set(l1+l2) - set(d0['cols_y']['desc'])))
    return d0

def get_comparison(
    df1: pd.DataFrame,
    cols_y: list,
    cols_index,
    coff_p: float=0.05,
    **kws,
    ):
    """
    Compare the x and y columns.
    
    Notes:
        Column types:
            cols_x: decrete and continuous
            cols_y: decrete and continuous

        Comparison types:
            1. continuous vs continuous -> correlation
            2. decrete vs continuous -> difference
            3. decrete vs decrete -> FE or chi square
    """
    ## 
    d1=get_cols_x_for_comparison(
    df1,
    cols_y,
    cols_index,
    **kws
    )
    from roux.lib.set import get_alt
    
    ## gather stats in a dictionary
    d2={}
    
    ## 1. correlations 
    from roux.stat.corr import get_corrs
    d2['correlation x vs y']=get_corrs(df1=df1,#.drop(['loci distance'],axis=1),
    method='spearman',
    cols=d1['cols_y']['cont'],
    cols_with=d1['cols_x']['cont'],#list(set(df1.columns.tolist())-set(['protein abundance difference (DELTA-WT)','responsiveness'])),
    # cols_with=['loci distance'],
                  coff_inflation_min=50,
             )
    if isinstance(coff_p,float):
        print(d2.keys())
        d2['correlation x vs y']=d2['correlation x vs y'].loc[(d2['correlation x vs y']['P']<0.05),:]
    
    ## 2. difference 
    from roux.stat.diff import get_diff
    for k in ['x','y']:
        d2[f"difference {k} vs {get_alt(['x','y'],k)}"]=get_diff(df1,
             cols_x=d1[f"cols_{k}"]['desc'],
             cols_y=d1[f"cols_{get_alt(['x','y'],k)}"]['cont'],
             cols_index=d1['cols_index'],
            )
        if isinstance(coff_p,float):
            d2[f"difference {k} vs {get_alt(['x','y'],k)}"]=(d2[f"difference {k} vs {get_alt(['x','y'],k)}"]
                                                            .loc[(d2[f"difference {k} vs {get_alt(['x','y'],k)}"]['P (MWU test)']<coff_p),:]
                                                            .sort_values('P (MWU test)')
                                                            )
    
    ## 3. association
    # df0=pd.DataFrame(itertools.product(d1['cols_y']['desc'],d1['cols_x']['desc'],)).rename(columns={0:'colx',1:'coly'},errors='raise')
    # df0.head(1)
    # from roux.lib.stat.diff import compare_classes
    # d2['association x vs y']=df0.join(df0.apply(lambda x: compare_classes(df1[x['colx']], df1[x['coly']]),axis=1).apply(pd.Series).rename(columns={0:'stat',1:'P'},errors='raise'))
    from roux.lib.stat.diff import compare_classes_many
    d2['association x vs y']=compare_classes_many(
        df1=df1,
        cols_y=d1['cols_y']['desc'],
        cols_x=d1['cols_x']['desc'],
        )
    if isinstance(coff_p,float):
        d2['association x vs y']=d2['association x vs y'].loc[(d2['association x vs y']['P']<coff_p),:].sort_values('P')
        
    ## rename to uniform column names
    d2['correlation x vs y']=(d2['correlation x vs y']
                            .rename(columns={'variable1':'variable x',
                                'variable2':'variable y',
                                '$r_s$':'stat',},
                                errors='raise')
                             .assign(**{'stat type':'$r_s$'})
                             )
    for k in ['x vs y','y vs x']:
        d2[f'difference {k}']=(d2[f'difference {k}']
                                .rename(columns={
                                    'P (MWU test)':'P',
                                    'stat (MWU test)':'stat',},
                                    errors='raise')
                                 .assign(**{'stat type':'MWU'})
                                 )
    d2['association x vs y']=(d2['association x vs y']
                            .rename(columns={'colx':'variable x',
                                'coly':'variable y',
                                            },
                                errors='raise')
                             .assign(**{'stat type':'FE'})
                             )
    ## gather
    return pd.concat(d2,axis=0,names=['comparison type'],
             ).reset_index(0)    



import pandas as pd
import numpy as np
from Bio import SeqIO,SeqRecord,Seq
from roux.lib.dfs import *
import logging

##defs
def reverse_complement(s): 
    """Reverse complement.

    Args:
        s (str): sequence

    Returns:
        s (str): reverse complemented sequence
    """
    return str((str2seq(s) if isinstance(s,str) else s).reverse_complement())

def fa2df(alignedfastap: str,ids2cols=False) -> pd.DataFrame:
    """_summary_

    Args:
        alignedfastap (str): path.
        ids2cols (bool, optional): ids of the sequences to columns. Defaults to False.

    Returns:
        DataFrame: output dataframe.
    """
    dtmp=pd.read_csv(alignedfastap,names=["c"],
                    header=None,)
    dtmp=dtmp.iloc[::2].reset_index(drop=True).join(dtmp.iloc[1::2].reset_index(drop=True),rsuffix='r')
    dtmp.columns=['id','sequence']
    dtmp=dtmp.set_index('id')
    dtmp.index=[i[1:] for i in dtmp.index]
    dtmp.index.name='id'
    if ids2cols:
        for i in dtmp.index:
            seqid,contig,strand,start,end=i.split('|')
            dtmp.loc[i,'seqid']=seqid
            dtmp.loc[i,'contig']=contig
            dtmp.loc[i,'strand']=strand
            dtmp.loc[i,'start']=start
            dtmp.loc[i,'end']=end
    return dtmp

def to_genomeocoords(genomecoord: str) -> tuple:
    """String-formated genome co-ordinates to separated values.

    Args:
        genomecoord (str):

    Raises:
        ValueError: format of the genome co-ordinates.

    Returns:
        tuple: separated values i.e. chrom,start,end,strand
    """
    try:
        chrom=genomecoord.split(':')[0]
    except:
        raise ValueError(genomecoord)
    start=genomecoord.split(':')[1].split('-')[0]

    end=genomecoord.split(':')[1].split('-')[1].replace('+','').replace('-','')

    tail=genomecoord.split(':')[1].replace(start,'')
    if tail.endswith('+'):
        strand='+'
    elif tail.endswith('-'):
        strand='-'
    else:
        strand=''
#     print(tail,strand)
    return chrom,start,end,strand

bed_colns=['chromosome', 'start', 'end', 'id', 'NM', 'strand']
def to_bed(df, col_genomeocoord):
    """Genome co-ordinates to bed.

    Args:
        df (DataFrame): input dataframe.
        col_genomeocoord (str): column with the genome coordinates.

    Returns:
        DataFrame: output dataframe.
    """
    df=df.dropna(subset=[col_genomeocoord])
    dbed=df.apply(lambda x: to_genomeocoords(x[col_genomeocoord]),axis=1).apply(pd.Series)
    if len(dbed)!=0:
        dbed.columns=['chromosome', 'start', 'end','strand']
        dbed['id']=df[col_genomeocoord]
        dbed['NM']=np.nan
        return dbed[bed_colns]
    else:
        return pd.DataFrame(columns=['chromosome', 'start', 'end','strand','id','NM'])

def to_seq(s : str,prt=False) -> Seq:
    """String to Seq object.

    Args:
        s (str): input string.
        prt (bool, optional): format. Defaults to False.

    Returns:
        Seq: Seq object.
    """
    if prt:
        alpha=Alphabet.ProteinAlphabet
    else:
        alpha=Alphabet.generic_dna
    return Seq.Seq(s,alpha)
    
def translate(dnaseq: str,fmtout=str,tax_id=None) -> str:
    """Translates a DNA sequence
    Args:
        dnaseq (str): DNA sequence
        fmtout (_type_, optional): format of output sequence. Defaults to str.
        tax_id (_type_, optional): _description_. Defaults to None.

    Returns:
        str: protein sequence.
    """
    if isinstance(dnaseq,str): 
        dnaseq=Seq.Seq(dnaseq)
    if tax_id is None:
        tax_id=1 # standard codon table. ref http://biopython.org/DIST/docs/tutorial/Tutorial.html#htoc25
    prtseq=dnaseq.translate(table=tax_id)
    if fmtout is str:
        return str(prtseq)
    else:
        return prtseq
    
## io file
### multiple seq fasta
def read_fasta(fap: str,key_type='id',duplicates=False) -> dict:
    """Read fasta

    Args:
        fap (str): path
        key_type (str, optional): key type. Defaults to 'id'.
        duplicates (bool, optional): duplicates present. Defaults to False.

    Returns:
        dict: data.

    TODOs:
        1. Check for duplicate keys.
    """
    if (not duplicates) or key_type=='id':
        try:
            id2seq=SeqIO.to_dict(SeqIO.parse(fap,format='fasta'))
            id2seq={k:str(id2seq[k].seq) for k in id2seq}
            return id2seq
        except:
            duplicates=True
    if duplicates or key_type=='description':
        id2seq={}
        for seq_record in SeqIO.parse(fap, "fasta"):
            id2seq[getattr(seq_record,key_type)]=str(seq_record.seq)
        return id2seq

def to_fasta(ids2seqs: dict,fastap: str,) -> str:
    """Save fasta file.

    Args:
        ids2seqs (dict): dictionary mapping the sequence name to the sequence.
        fastap (str): path of the fasta file

    Returns:
        fastap (str): path of the fasta file
    """
    from roux.lib.sys import makedirs,exists
    # if exists(fastap) and not force: return
    makedirs(fastap)
    seqs = (SeqRecord.SeqRecord(Seq.Seq(ids2seqs[id]), id) for id in ids2seqs)
    SeqIO.write(seqs, fastap, "fasta")
    return fastap

def dedup_fasta(fap: str,faoutp=None) -> str:
    """Deduplicate fasta file.

    Args:
        fap (str): path
        faoutp (_type_, optional): output path. Defaults to None.

    Returns:
        str: output path.
    """
    return to_fasta(read_fasta(fap,key_type='description'),
             fastap=f"{splitext(fap)[0]}_dedup.fasta" if faoutp is None else faoutp)    
    


import pandas as pd
import matplotlib.pyplot as plt
## TODOs: use ## stats

def plot_workflow_log(
    dplot: pd.DataFrame
    ) -> plt.Axes:
    """Plot workflow log.

    Args:
        dplot (pd.DataFrame): input data (dparam).

    Returns:
        plt.Axes: output.

    TODOs:
        1. use the statistics tagged as `## stats`. 
    """
    parameters_count_max=dplot.groupby(['function name']).agg({'parameter name input list':lambda x: len(x)+1}).max().values[0]
    plt.figure(figsize=[parameters_count_max*1.5,#*0.3,
                        len(dplot)*0.5+2,])
    ax=plt.subplot(1,5,2)
    #     ax=plt.subplot()
    from roux.viz.colors import saturate_color
    elements=[
                'script',
                'function',
    ]
    for elementi,element in enumerate(elements):
        _=dplot.apply(lambda x: ax.text(x[f"{element} x{''  if element!='parameter' else ' input'}"],
                                        x[f"{element} y{''  if element!='parameter' else ' input'}"],
                                        x[f"{element} name{''  if element!='parameter' else ' input'}"]),axis=1)
    _=dplot.apply(lambda x: ax.annotate(x["parameter name output"],
                xy=(x['parameter x input'], x['parameter y input']), xycoords='data',
                xytext=(x['parameter x output'], x['parameter y output']), 
                textcoords='data',
    #             size=20, 
                va="center", ha="center",
                arrowprops=dict(arrowstyle='<|-',alpha=0.5,color='lime',lw=4,
                                connectionstyle="arc3,rad=0.4"),
                ),axis=1)    
    ax.set_ylim(len(dplot),0)
    #     ax.set_xlim(0,parameters_count_max)
    ax.set_axis_off()                                                
    return ax 


import pandas as pd

def to_class(cls):
    """Get the decorator to attach functions. 
    
    Parameters:
        cls (class): class object.

    Returns:
        decorator (decorator): decorator object.
    
    References:
        https://gist.github.com/mgarod/09aa9c3d8a52a980bd4d738e52e5b97a
    """
    from functools import wraps
    def decorator(func):
        @wraps(func) 
        def wrapper(self, *args, **kwargs): 
            return func(self._obj,*args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        return func
    return decorator

@pd.api.extensions.register_dataframe_accessor("rd")
class rd:
    """`roux-dataframe` (`.rd`) extension.
    """
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

# create the `roux-dataframe` (`.rd`) decorator
to_rd=to_class(rd)

# @pd.api.extensions.register_dataframe_accessor("stat")
# class stat:
#     def __init__(self, pandas_obj):
#         self._obj = pandas_obj





from roux.lib.sys import isdir,exists,dirname,basename,makedirs,basenamenoext,info,logging
from roux.lib.str import replace_many
from roux.lib.io import to_path
from roux.lib.set import unique
import pandas as pd

def get_quoted_path(
    s1:str
    ) -> str:
    """Quoted paths.

    Args:
        s1 (str): path.

    Returns:
        str: quoted path.
    """
    s1=f'"{s1}"'
    if "{metadata" in s1:
        s1='f'+s1
        s1=replace_many(s1,{"[":"['","]":"']"})
    return s1

def get_path(
    s: str,
    validate: bool,
    prefixes=['data/','metadata/','plot/'],
    test=False) -> str:
    """Extract pathsfrom a line of code.

    Args:
        s (str): line of code.
        validate (bool): validate the output.
        prefixes (list, optional): allowed prefixes. Defaults to ['data/','metadata/','plot/'].
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: path.

    TODOs:
        1. Use wildcards i.e. *'s.
    """
    if ('=' in s):# and ((not "='" in s) or (not '="' in s)):
        s=s[s.find('=')+1:]
    if ')' in s:
        s=s.split(')')[0]
    l0=[s_ for s_ in s.split(',') if '/' in s_]
    l0=[dirname(s.split('{')[0]) if (('{' in s) and (not "{metadata" in s)) else s for s in l0]
    l0=[dirname(s.split('*')[0]) if '*' in s else s for s in l0]
    if test:info(l0)
    if len(l0)!=1:
        if validate:
            assert(len(l0)==1)
        else:
            s1=''
    else:
        s=l0[0]
        s=replace_many(s,['read_table','to_table',
                              "read_plot",'to_plot',
                              ' ',"f'",'f"','"',"'","(",")"],
                           '',
                          ignore=True)
        # if test:info(s)
        if any([s.startswith(s_) for s_ in prefixes]):
            s1=s
        else:
            s1=''
    if test:info(s1)
    s1=get_quoted_path(s1)
    s1=replace_many(s1,{'//':'/','///':'/'},ignore=True)
    return s1

def remove_dirs_from_outputs(
    outputs: list,
    test: bool=False
    ) -> list:
    """Remove directories from the output paths.

    Args:
        outputs (list): output paths.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        list: paths.
    """
    l_=[s.replace('"','') for s in outputs if not s.startswith('f')]
    if any([isdir(p) for p in l_]) and any([not isdir(p) for p in l_]):
        # if filepath is available remove the directory paths
        outputs=[f'"{p}"' for p in l_ if not isdir(p)]
        if test: logging.info("directory paths removed")
    if test: print(outputs)
    return outputs

def get_ios(
    l: list,
    test=False) -> tuple:
    """Get input and output (IO) paths.

    Args:
        l (list): list of lines of code.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        tuple: paths of inputs and outputs.
    """
    ios=[s_ for s_ in l if (('data/' in s_) or ('plot/' in s_) or ('figs/' in s_))]
    if test:info(ios)
    inputs=[f'{get_path(s,validate=False,test=test)}' for s in ios if ('read_' in s) or (s.lstrip().startswith('p='))]
    outputs=[f'{get_path(s,validate=False,test=test)}' for s in ios if (('to_' in s) or (s.lstrip().startswith('outp='))) and (not 'prefix' in s)]
    outputs=remove_dirs_from_outputs(outputs,test=test)
    inputs,outputs=[p for p in inputs if p!='""'],[p for p in outputs if p!='""']
    return unique(inputs),unique(outputs)

def get_name(
    s : str, 
    i: int,
    sep_step: str='## step',
    ) -> str:
    """Get name of the function.

    Args:
        s (str): lines in markdown format.
        sep_step (str, optional): separator marking the start of a step. Defaults to "## step".        
        i (int): index of the step.

    Returns:
        str: name of the function.
    """
    assert s.startswith(f'# {sep_step}')
    assert s.count(sep_step)==1
    s1=to_path(s.replace(f'# {sep_step}',f'step{i:02}')).lower().replace('/','_')
    s1=s1 if len(s1)>=80 else s1[:80]
    s1=replace_many(s1,{' ':'_','.':'_'},ignore=True)
    return s1

def get_step(
    l: list,
    name: str,
    sep_step: str='## step',
    sep_step_end: str='## tests',
    test=False,
    tab='    ') -> dict:
    """Get code for a step.

    Args:
        l (list): list of lines of code
        name (str): name of the function.
        test (bool, optional): test mode. Defaults to False.
        tab (str, optional): tab format. Defaults to '    '.

    Returns:
        dict: step name to code map. 
    """
    # to_fun():
    if test:info(name,l[-1])
    docs=[]
    code_with_comments=[]
    get_docs=True
    get_code=False
    ## in markdown cell
    for i,s in enumerate(l):
        if (s.startswith('#') and not s.startswith('# In[')) and get_docs:
            if i==0:
                # info(s)
                s='# '+s.split(sep_step.split('## ')[0]+' ')[1][0].upper()+s.split(sep_step.split('## ')[0]+' ')[1][1:]+'.'
            docs.append(s)
        else:
            get_docs=False
            get_code=True
        if get_code:
            code_with_comments.append(s)
            
    if test:info(docs)
    ## [X: gets commented code] steps marked as '## #1', '## #2' ..
    # docs+=[s for s in l if s.startswith('## #')]
    ## remove leading #
    docs='\n'.join([s[2:] for s in docs])
    
    code=[s for s in code_with_comments if not s.startswith('#')]
#     info(code)
    inputs,outputs=get_ios(code,test=test)
    if len(set(inputs) & set(outputs)) !=0:
        inputs=[s for s in inputs if not s in outputs]
#     if name.endswith('network_networkx'):
#         info(inputs,outputs)
    if test:info(inputs,outputs)
    if 'plot' in name:
        output_type='plots'
    elif 'figure' in name:
        output_type='figures'
    else:
        output_type='data'
    inputs_str=f',\n{tab*3}'.join(inputs)
    outputs_str=f',\n{tab*3}'.join(outputs)
    if output_type!='data' and (not any([isdir(p.replace('"','')) for p in outputs])):
        outputs_str=f'report([\n{tab*3}{outputs_str}],\n{tab*3}category="{output_type}")'
#     elif output_type in ['plots','figures']:
#         outputs_str=''
    ## snakemake rule
    config=[
    f"    rule {name}:",
    f"        input:",
    f"            {inputs_str}",
    f"        output:",
    f"            {outputs_str}",
    f"        run:",
    f"            from lib.{name.split('_step')[0]} import {name}",
    f"            {name}(metadata=metadata)",
    ]
    config_str='\n'.join(config)+'\n'
    quotes='"""'
    function=[
    f"def {name}(metadata=None):",
    f"    {quotes}{docs}",
    f"",
    "    Parameters:",
    "        metadata (dict): Dictionary containing information required for the analysis. Metadata files are located in `metadata` folder are read using `roux.workflow.io.read_metadata` function.",
    f"",
    f"    Snakemake rule:",
    '\n'.join([tab+s+tab for s in config]),
    f"    {quotes}",
    "    "+'\n    '.join(code_with_comments),
    ]
    if test:info(function[0])
    function='\n'.join(function)
    return {'function':function,'config':config_str,'inputs':inputs,'outputs':outputs,}

def to_task(
    notebookp,
    task=None,
    sep_step: str='## step',
    sep_step_end: str='## tests',
    notebook_suffix: str='_v',
    force=False,
    validate=False,
    path_prefix=None,
    verbose=True,
    test=False) -> str:
    """Get the lines of code for a task (script to be saved as an individual `.py` file).

    Args:
        notebookp (_type_): path of the notebook.
        sep_step (str, optional): separator marking the start of a step. Defaults to "## step".        
        sep_step_end (str, optional): separator marking the end of a step. Defaults to "## tests".        
        notebook_suffix (str, optional): suffix of the notebook file to be considered as a "task".
        force (bool, optional): overwrite output. Defaults to False.
        validate (bool, optional): validate output. Defaults to False.
        path_prefix (_type_, optional): prefix to the path. Defaults to None.
        verbose (bool, optional): show verbose. Defaults to True.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: lines of the code.
    """
    if not sep_step.startswith('## '):
        raise ValueError(f"{sep_step} should start with '## '")
    # from roux.lib.str import removesuffix
    if notebook_suffix!='':
        pyp=f"{dirname(notebookp)}/lib/task{basenamenoext(notebookp).split(notebook_suffix)[0] if task is None else task}.py"
    else:
        pyp=f"{dirname(notebookp)}/lib/task{basenamenoext(notebookp) if task is None else task}.py"
    # info(notebookp,pyp)
    # brk
    if exists(pyp) and not force and not test: return 
    if verbose: info(basename(notebookp))
    from roux.workflow.io import to_py, get_lines
    if not test:
        to_py(notebookp,
             pyp=pyp.replace('/lib/','/.lib/'),
             force=force)
    l0=get_lines(notebookp, keep_comments=True)
    # print(l0)
    # info(sep_step,sep_step_end)
    if not path_prefix is None:
        l0=[replace_many(s,{'data/':'../data/'},ignore=True) for s in l0]
    taskn=basenamenoext(pyp)
    if test: info(taskn)
    d0={}
    get=False
    get_header=True
    l1=[] # start of the file
    l2=[] # start of code
    for s in l0:
        if s.startswith(f'# {sep_step}'):
            l2=[] # start of code
            get_header=False        
            get=True
            k=s
        elif s.startswith(sep_step_end) and len(l2)!=0:
            get=False
            stepn=f"{taskn}_{get_name(k,i=len(d0.keys())+1,sep_step=sep_step)}"
            d0[stepn]=get_step(l2,name=stepn,
                               sep_step=sep_step,
                               sep_step_end=sep_step_end,
                               test=test or verbose,
                              )
            l2=[]
            stepn=None
        elif s.startswith('# In'):
            continue
        elif get_header:
            if 'get_ipython' in s:
                continue
            l1.append(s)
        if get:
            l2.append(s)
    l3=[]
    for s in l1:
        l3.append(s)
        header=s.startswith('#')
        if (not 'import' in s and not header) or s.startswith('# ## '):
            break
    df0=pd.DataFrame(d0).T
    df0.index.name='step name'
    df0=df0.reset_index()
    df0.index.name='step #'
    df0=df0.reset_index()
    if len(df0)==0 and not validate:
        # if verbose: 
        logging.warning('no functions found')
        return None,None 
    if not test:
        makedirs(pyp)
        with open(pyp,'w') as f:
            f.write('\n'.join(l3)+'\n\n'+'\n\n'.join(df0['function'].tolist()))
    return '\n'.join(df0['config'].tolist()).replace('\n    ','\n').replace('    rule','rule'),df0['outputs'].tolist()



def git_commit(
    repop: str,
    suffix_message: str=''):
    """Version control.

    Args:
        repop (str): path to the repository.
        suffix_message (str, optional): add suffix to the version (commit) message. Defaults to ''.
    """
    from git import Repo
    repo=Repo(repop,search_parent_directories=True)
    def commit_changes(repo):
        """if any"""
        repo.git.add(update=True)
        repo.index.commit('auto-update'+suffix_message)

    if len(repo.untracked_files)!=0:
        from roux.lib.sys import input_binary
        print(len(repo.untracked_files),'untracked file/s in the repo:',repo.untracked_files)
        yes=input_binary("add all of them? [y/n]")
        if yes:
            repo.git.add(repo.untracked_files)
        else:
            yes=input_binary("add none of them? [y/n]")
            if yes:
                commit_changes(repo)
            else:
                repo.git.add(eval(input('list of files in py syntax:')))
        repo.index.commit('manual-update'+suffix_message)
    else:
        commit_changes(repo)
    print('git-committed')
    
    


from roux.global_imports import *

# set enrichment
def get_enrichment(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    background: int,
    colid: str='gene id',
    colref: str='gene set id',
    colrefname: str='gene set name',
    colreftype: str='gene set type',
    colrank: str='rank',
    outd: str=None,
    name: str=None,
    # enrichr
    cutoff: float=0.05, # only used for plotting. 
    # prerank
    permutation_num: int=1000,
    # etc
    verbose: bool=False,
    no_plot: bool=True,
    **kws_prerank,
    ):
    """Get enrichments between sets.

    Args:
        df1 (pd.DataFrame): test data.
        df2 (pd.DataFrame): reference set data.
        background (int): background size.
        colid (str, optional): column containing unique ids of the elements. Defaults to 'gene id'.
        colref (str, optional): column containing the unique ids of the sets. Defaults to 'gene set id'.
        colrefname (str, optional): column containing names of the sets. Defaults to 'gene set name'.
        colreftype (str, optional): column containing the type/group name of the sets. Defaults to 'gene set type'.
        colrank (str, optional): column containing the ranks. Defaults to 'rank'.
        outd (str, optional): output directory path. Defaults to None.
        name (str, optional): name of the result. Defaults to None.
        cutoff (float, optional): p-value cutoff. Defaults to 0.05.
        verbose (bool, optional): verbose. Defaults to False.
        no_plot (bool, optional): do no plot. Defaults to True.

    Returns:
        pd.DataFrame: if rank -> high rank first within the leading edge gene ids.
        
    Notes:
        1. Unique ids are provided as inputs.
    """
    import gseapy as gp
    name=get_name(df2,colreftype)
    
    import tempfile
    with tempfile.TemporaryDirectory() as p:
        outd=outd if not outd is None else p
        o1 = gp.enrichr(
            gene_list=df1[colid].unique().tolist(),
             # or gene_list=glist
             description=name,
             gene_sets=df2.rd.to_dict([colref,colid]),
             background=background, # or the number of genes, e.g 20000
             outdir=to_path(f'{outd}/{name}'),
             cutoff=cutoff, # only used for plotting.
             verbose=verbose,
             no_plot=no_plot,
             # **kws,
             )
        df3=o1.results
    if len(df3)==0:
        logging.error('aborting because no enrichments found.')
        return
    df3=df3.rename(columns={'Term':colref,
                             'P-value':'P (FE test)',
                             'Adjusted P-value':'P (FE test, FDR corrected)',
                             'Genes':f'{colid}s overlap'},errors='raise')
    df_=df3['Overlap'].str.split('/',expand=True).rename(columns={0:f"{colid}s overlap size",
                                                         1:f"{colid}s per {colref}"}).applymap(int)
    df3=df3.join(df_)
#     df3['overlap %']=df3['Overlap'].apply(eval)*100    
    df3['overlap %']=df3.apply(lambda x: (x[f"{colid}s overlap size"]/x[f"{colid}s per {colref}"])*100,axis=1)
    df3=df3.drop(['Gene_set','Overlap',],axis=1)
    info('enrichr: '+perc_label(sum(df3['P (FE test, FDR corrected)']<=0.05),len(df3)))
    if not colrank in df1:
        if colrefname in df2:
            df3[colrefname]=df3[colref].map(df2.set_index(colref)[colrefname].drop_duplicates())
        return df3
    if df3[f"{colid}s overlap size"].max()<2:
        logging.error("df3[f'{colid}s overlap size'].max()<2 # can not run prerank")
        return df3
    with tempfile.TemporaryDirectory() as p:
        outd=outd if not outd is None else p
        o2 = gp.prerank(rnk=df1.loc[:,[colid,colrank]],
                         gene_sets=df2.rd.to_dict([colref,colid]),
                         min_size=2,
                         max_size=df3[f"{colid}s overlap size"].max(),
                         processes=1,
                         permutation_num=permutation_num, # reduce number to speed up testing
                         ascending=False, # look for high number 
                         outdir=to_path(f'{outd}/{name}'),
                         pheno_pos='high',
                         pheno_neg='low',
                         format='png',
                        no_plot=no_plot,
                        seed=1,
                        graph_num=1,
                        **kws_prerank,
                       )
        to_dict({'prerank':o2},
               f'{outd}/{name}/prerank.joblib')
        df4=o2.res2d.reset_index()
#     es	nes	pval	fdr	geneset_size	matched_size	genes	ledge_genes
    df4=df4.rename(columns={
        'Term':colref,
        'es':'enrichment score',
        'nes':'normalized enrichment score', 
        'pval':'P (GSEA test)',
        'fdr':'FDR (GSEA test)',
        'matched_size':f"{colid}s overlap size",
        'geneset_size':f"{colid}s per {colref}",
        'ledge_genes':f'{colid}s leading edge',
        },
        errors='raise')
    df4['overlap %']=df4.apply(lambda x: (x[f"{colid}s overlap size"]/x[f"{colid}s per {colref}"])*100,axis=1)
    df4=df4.drop(['genes',],axis=1)
    info('preraked: '+perc_label(sum(df4['P (GSEA test)']<=0.05),len(df4)))
    
    df5=df3.merge(df4,
              on='gene set id',
              how='left',
             validate="1:1",
                 suffixes=['',' (prerank)'])
    df5=df5.drop([
#                 f'{colid}s',
                  f'{colid}s per {colref} (prerank)',f'{colid}s overlap size (prerank)','overlap % (prerank)'],axis=1)
    return df5

def get_enrichments(df1: pd.DataFrame,
                    df2: pd.DataFrame,
                    background: int,
                    coltest: str='subset',
                    colid: str='gene id',
                    colref: str='gene set id',
                    colreftype: str='gene set type',
                    fast: bool=False,
                    **kws) -> pd.DataFrame:
    """Get enrichments between sets, iterate over types/groups of test elements e.g. upregulated and downregulated genes.

    Args:
        df1 (pd.DataFrame): test data.
        df2 (pd.DataFrame): reference set data.
        background (int): background size.
        colid (str, optional): column containing unique ids of the elements. Defaults to 'gene id'.
        colref (str, optional): column containing the unique ids of the sets. Defaults to 'gene set id'.
        colrefname (str, optional): column containing names of the sets. Defaults to 'gene set name'.
        colreftype (str, optional): column containing the type/group name of the sets. Defaults to 'gene set type'.
        fast (bool, optional): parallel processing. Defaults to False.

    Returns:
        pd.DataFrame: output.
    """

    return getattr(df1.groupby(coltest),'progress_apply' if not fast else 'parallel_apply')(lambda df_: df2.groupby(colreftype).apply(lambda df: get_enrichment(df_,df,background=background,**kws)).reset_index(0)).reset_index(0)
    


import pandas as pd
import logging
"""
pip install --upgrade google-api-python-client google-auth-httplib2 google-auth-oauthlib
"""

def get_service(service_name='drive',access_limit=True,client_config=None):
    """
    Creates a google service object. 
    
    :param service_name: name of the service e.g. drive
    :param access_limit: True is access limited else False
    :param client_config: custom client config
    ...
    :return: google service object
    
    Ref:
    https://developers.google.com/drive/api/v3/about-auth
    """
    service_name2params={
    'drive': {
        'scope':'https://www.googleapis.com/auth/drive.readonly',
        'build':'drive',
        'version': 'v3'},        
    'slides': {
        'scope':'https://www.googleapis.com/auth/presentations.readonly',
        'build':'slides',
        'version': 'v1'},
    'sheets':{
        'scope':'https://www.googleapis.com/auth/spreadsheets.readonly',
        'build':'sheets',
        'version': 'v4',},
    'customsearch':{
        'build':'customsearch',
        'version': 'v1',
        'scope':'https://www.googleapis.com/auth/cse',
    },
    }
    if client_config is None:
        from getpass import getpass
        client_config=eval(getpass())
    from googleapiclient.discovery import build
    from google_auth_oauthlib.flow import InstalledAppFlow
#     SCOPES=[]
#     if 'scope' in service_name2params[service_name]:
#     SCOPES = SCOPES.append([service_name2params[service_name]['scope']])
    SCOPES = [service_name2params[service_name]['scope']]
    if not access_limit:
        SCOPES = [s.replace('.readonly','') for s in SCOPES]
    logging.warning(SCOPES)
    flow = InstalledAppFlow.from_client_config(client_config,SCOPES)
    creds = flow.run_console(port=0)
    service = build(service_name,service_name2params[service_name]['version'], credentials=creds,
                   cache_discovery=False)
    return service

get_service_drive=get_service
    
def list_files_in_folder(service,folderid,
                         filetype=None,
                         fileext=None,                         
                         test=False):
    """
    Lists files in a google drive folder.
    
    :param service: service object e.g. drive
    :param folderid: folder id from google drive
    :param filetype: specify file type
    :param fileext: specify file extension
    :param test: True if verbose else False
    ...
    :return: list of files in the folder     
    """
    filetype2mimetype={"audio":"application/vnd.google-apps.audio", #
    "document":"application/vnd.google-apps.document", #Google Docs
    "drive":"application/vnd.google-apps.drive-sdk", #3rd party shortcut
    "drawing":"application/vnd.google-apps.drawing", #Google Drawing
    "file":"application/vnd.google-apps.file", #Google Drive file
    "folder":"application/vnd.google-apps.folder", #Google Drive folder
    "form":"application/vnd.google-apps.form", #Google Forms
    "fusiontable":"application/vnd.google-apps.fusiontable", #Google Fusion Tables
    "map":"application/vnd.google-apps.map", #Google My Maps
    "photo":"application/vnd.google-apps.photo", #
    "presentation":"application/vnd.google-apps.presentation", #Google Slides
    "script":"application/vnd.google-apps.script", #Google Apps Scripts
    "shortcut":"application/vnd.google-apps.shortcut", #Shortcut
    "site":"application/vnd.google-apps.site", #Google Sites
    "spreadsheet":"application/vnd.google-apps.spreadsheet", #Google Sheets
    "unknown":"application/vnd.google-apps.unknown", #
    "video":"application/vnd.google-apps.video",}

    results = service.files().list(
        q=f"'{folderid}' in parents"+(f" and mimeType='{filetype2mimetype[filetype]}'" if not filetype is None else ""),
        fields="nextPageToken, files(id, name)",).execute()
    items = results.get('files', [])
    name2id={d['name']:d['id'] for d in items}
    if not items:
        print(items)
        print('No files found.')
    else:
        if test:
            print(name2id)
    if not fileext is None:
        name2id={k:name2id[k] for k in name2id if k.endswith(fileext)}
    return name2id
def get_file_id(p):return p.split('/')[5]
def download_file(p=None,
                  file_id=None,
                  service=None,
                  outd=None,
                  outp=None,
                  convert=False,
                  force=False,
                  test=False):
    """
    Downloads a specified file.
    
    :param service: google service object
    :param file_id: file id as on google drive
    :param filetypes: specify file type
    :param outp: path to the ouput file
    :param test: True if verbose else False
    
    Ref: https://developers.google.com/drive/api/v3/ref-export-formats
    """
    if not p is None:
        if file_id is None:
            file_id=get_file_id(p)
        else:
            raise ValueError("define p or file_id")
    if not outp is None:
        if exists(outp) and not force: return outp
    from googleapiclient.http import MediaIoBaseDownload
    from roux.lib.sys import makedirs
    from os.path import dirname,exists
    import io
    if service is None:
        service=get_service_drive()
    if not outd is None or not convert:
        file = service.files().get(fileId=file_id).execute()
    else:
        file={'name':'tmp'}
    if (outp is None) and (not outd is None):
        outp=f"{outd}/{file['name']}"
        if exists(outp) and not force: return outp
    import tempfile
    with tempfile.TemporaryDirectory() as outd_:
        if not outd is None:
            outd_=outd
        else:
            outd=outd_
        outp=f"{outd}/{file['name']}"
        if not convert:
            request = service.files().get_media(fileId=file_id)
        else:
            request = service.files().export_media(fileId=file_id,mimeType=file['mimeType'])
        makedirs(outp)
        fh = io.FileIO(file=outp,mode='w')
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            if test:
                print( f"Downloading {outp}: {int(status.progress() * 100)}")
        return outp

def upload_file(service,filep,folder_id,test=False):
    """
    Uploads a local file onto google drive.
    
    :param service: google service object
    :param filep: path of the file
    :param folder_id: id of the folder on google drive where the file will be uploaded 
    :param test: True is verbose else False
    ...
    :return: id of the uploaded file 
    """
    from googleapiclient.http import MediaFileUpload
    from os.path import basename
    file_metadata = {'name': basename(filep),
                    'parents': [folder_id]}
    media = MediaFileUpload(filep, mimetype=f"image/{filep.split('.')[1]}")
    file_name2id=list_files_in_folder(service,folderid=folder_id,filetype=None,test=False)
    if not basename(filep) in file_name2id:
        file = service.files().create(body=file_metadata,
                                        media_body=media,
                                        fields='id').execute()
        file_id=file['id']
    else:
        file_id=file_name2id[basename(filep)]
        if test:
            print(f"replacing {basename(filep)} in the folder")
        file=service.files().get(fileId=file_name2id[basename(filep)]).execute()
        del file['id']
        response = service.files().update(fileId=file_id,
                                          body=file,
                                          media_body=media,
                                         ).execute()
    return file_id

def upload_files(service,ps,folder_id,**kws):
    from tqdm import tqdm
    l1=[]
    for p in tqdm(ps):
        l1.append(upload_file(service,p,folder_id,**kws))
    return l1

def download_drawings(folderid,outd,service=None,test=False):
    """
    Download specific files: drawings
    
    TODOs:
    1. use download_file
    """
    if service is None:
        service=get_service_drive()
    filename2id=list_files_in_folder(service=service,
                              filetype='drawing',
                              folderid=folderid,
                              test=test,)           
    for n in filename2id:
        download_file(service,filename2id[n],[
#                                                 'image/png',
                                              'image/svg+xml'],
                      f"{outd}/{n}",test=test)

class slides():
    def get_page_ids(service,presentation_id):
        presentation = service.presentations().get(
            presentationId=presentation_id).execute()
        slides = presentation.get('slides')
        logging.info(f'total slides={len(slides)}')
        return [slide.get('objectId') for i, slide in enumerate(slides)]

    def create_image(service, presentation_id, page_id,image_id):
        """
        image less than 1.5 Mb
        """
        import numpy as np
        IMAGE_URL = (f'https://drive.google.com/uc?id={image_id}')
        image_id = f'image_{np.random.rand()}'.replace('.','')
        size = {
            'magnitude': 576,
            'unit': 'PT'
        }
        requests = []
        requests.append({
            'createImage': {
                'objectId': image_id,
                'url': IMAGE_URL,
                'elementProperties': {
                    'pageObjectId': page_id,
                    'size': {
                        'height': size,
                        'width': size
                    },
                    'transform': {
                        'scaleX': 1,
                        'scaleY': 1,
                        'unit': 'EMU'
                    }
                }
            }
        })
        # Execute the request.
        body = {
            'requests': requests
        }
        response = service.presentations() \
            .batchUpdate(presentationId=presentation_id, body=body).execute()
        create_image_response = response.get('replies')[0].get('createImage')
#         print('Created image with ID: {0}'.format(
#             create_image_response.get('objectId')))

        # [END slides_create_image]
        return create_image_response.get('objectId')
#     def update_images(presentation_id,page_id2image_id):
#         create_image(service, presentation_id, page_id,image_id)
#         page_ids=get_page_ids(service,presentation_id)
#         zip(page_ids)


def get_comments(fileid,
                 fields='comments/quotedFileContent/value,comments/content,comments/id',
                service=None):
    """
    Get comments.
    
        fields: comments/
                    kind:
                    id:
                    createdTime:
                    modifiedTime:
                    author:
                        kind:
                        displayName:
                        photoLink:
                        me:
                            True
                    htmlContent:
                    content:
                    deleted:
                    quotedFileContent:
                        mimeType:
                        value:
                    anchor:
                    replies:
                        []
    """
    def apply_(service,fileId,fields,**kws_list):
        comments = service.comments().list(fileId=fileId,fields=fields,**kws_list).execute()
        df1=pd.DataFrame(pd.concat({di:pd.Series({k:d[k] for k in d}) for di,d in enumerate(comments['comments'])},
                 axis=0)).reset_index().rename(columns={'level_0':'comment #',
                                                        'level_1':'key',
                                                        0:'value'})
        df1['value']=df1['value'].apply(lambda x: ','.join(x.values()) if isinstance(x,dict) else x)
        df1=df1.set_index(['comment #','key'])
        df1=df1.unstack(1).droplevel(0,1)
        df1['link']=df1['id'].apply(lambda x: f"https://drive.google.com/file/d/{fileId}/edit?disco={x}")
        df1=df1.rename(columns={'content':'comment',
                                'quotedFileContent':'text'}).drop(['id'],axis=1)
        return df1
    if service is None: service=get_service()    
    if isinstance(fileid,str):
        fileid=[fileid]
    df1=pd.concat({k:apply_(service,fileId=k,
                    #fields='comments',
                     fields=fields,# nextPageToken',
                     includeDeleted='false',
                     pageSize=100) for k in fileid},
             axis=0)
    return df1

def search(query,results=1,
           service=None,
           **kws_search):
    """
    Google search.
    
    :param query: exact terms
    ...
    :return: dict
    """
    if service is None: service=get_service("customsearch")
    # https://developers.google.com/custom-search/v1/reference/rest/v1/cse/list
    return service.cse().list(
      exactTerms=query,
      cx='46377b0459c06e668',
      num=results,
        **kws_search
    ).execute()
#     res

def get_search_strings(text,num=5,test=False):
    """
    Google search.
    
    :param text: string
    :param num: number of results
    :param test: True if verbose else False
    ...
    :return lines: list
    """
    lines=text.split("\n")
    lines=[s.split('.')[0].split(';')[0].strip() for s in lines]    
    lines=sorted(lines, key=len)[::-1][:num+2]
    import unicodedata
    lines=[unicodedata.normalize("NFKD", s).strip() for s in lines]
    if test: print(lines)
    cs=[sum([c.isalpha() for c in s])/len(s) for s in lines]
    lines=[x for _,x in sorted(zip(cs,lines))][::-1][:num]
    if test: print(lines)
    return lines

def get_metadata_of_paper(file_id,service_drive,service_search,
                          metadata=None,
                          force=False,
                         test=False):
    """
    Get the metadata of a pdf document.
    """
    import numpy as np
    if metadata is None: 
        metadata={'queries':np.nan,
                                  'titles':np.nan,
                                  'data':np.nan}
    else:
        if (not pd.isnull(metadata['data'])) and (not force):
            return metadata        
    if pd.isnull(metadata['queries']) or force:
        content = service_drive.files().get_media(fileId=file_id).execute()
        from roux.lib.io.text import pdf_to_text
        text=pdf_to_text(pdf_path=content,
                              pages=[2,3])
        metadata['queries']=get_search_strings(text,num=3,test=False)
    if test: print(metadata['queries'])
    if pd.isnull(metadata['titles']) or force:
        metadata['titles']=[]
        for k in metadata['queries']:
            res=search(query=k,results=1,
                       service=service_search,
            #            **kws_search
                  )
            if not 'items' in res:
                continue
            try:
                title=res['items'][0]['pagemap']['metatags'][0]['og:title']
            except:
                title=[d['htmlTitle'] for d in res['items']][0]
            metadata['titles'].append(title)
    if test: print(metadata['titles'])
    if pd.isnull(metadata['data']) or force:
        for k in metadata['titles']:
            from scholarly import scholarly
            d=scholarly.search_pubs(k)
            try:
                metadata['data']=next(d)
            except:
                continue
    return metadata

def share(drive_service,
          content_id,
          share=False,
          unshare=False,
         user_permission=None,
         permissionId='anyoneWithLink',
         ):
    """
    :params user_permission:     
    user_permission = {
        'type': 'anyone',
        'role': 'reader',
        'email':'@'
    }
    Ref:
    https://developers.google.com/drive/api/v3/manage-sharing
    """
    def callback(request_id, response, exception):
        if exception:
            print (exception)
        else:
            if hasattr(response,'get'):
                return response.get('id')
            else:
                print(response)
    if sum([share,unshare])!=1:
        logging.warning('either share or unshare.')
        return 
    batch = drive_service.new_batch_http_request(callback=callback)
    if share:
        batch.add(drive_service.permissions().create(
                fileId=content_id,
                body=user_permission,
                fields='id',
        ))
        return batch.execute()
    if unshare:
        batch.add(drive_service.permissions().delete(
                fileId=content_id,
                fields='id',
                permissionId=permissionId,  
        ))
        return batch.execute()


from argparse import ArgumentError
import pandas as pd
import numpy as np
import scipy as sc
import matplotlib.pyplot as plt
import logging
from icecream import ic as info
from roux.lib.set import *

# difference between values
def get_ratio_sorted(a: float,b: float,increase=True) -> float:
    """Get ratio sorted.

    Args:
        a (float): value #1.
        b (float): value #2.
        increase (bool, optional): check for increase. Defaults to True.

    Returns:
        float: output.
    """
    l=sorted([a,b])
    if not increase:
        l=l[::-1]
    if l[0]!=0 and l[1]!=0:
        return l[1]/l[0]

def diff(a: float,b: float,absolute=True) -> float:
    """Get difference

    Args:
        a (float): value #1.
        b (float): value #2.
        absolute (bool, optional): get absolute difference. Defaults to True.

    Returns:
        float: output.
    """
    diff=a-b
    if absolute:
        return abs(diff)
    else:
        return diff

def get_diff_sorted(a: float,b: float) -> float:
    """Difference sorted/absolute.

    Args:
        a (float): value #1.
        b (float): value #2.

    Returns:
        float: output.
    """
    return diff(a,b,absolute=True)

def balance(a: float,b: float,absolute=True) -> float:
    """Balance.

    Args:
        a (float): value #1.
        b (float): value #2.
        absolute (bool, optional): absolute difference. Defaults to True.

    Returns:
        float: output.
    """
    sum_=a+b
    if sum_!=0:
        return 1-(diff(a,b,absolute=absolute)/(sum_))
    else:
        return np.nan

# differnece by groups
def get_col2metrics(df,
                    colxs: list,
                    coly: str,
                    method='mannwhitneyu',
                    alternative='two-sided') -> dict:
    """Get column-wise metrics.

    Args:
        df (DataFrame): input dataframe.
        colxs (list): columns.
        coly (str): y column, contains 2 values.
        method (str, optional): method name. Defaults to 'mannwhitneyu'.
        alternative (str, optional): alternative for the statistical test. Defaults to 'two-sided'.

    Returns:
        dict: output.
    """
    from scipy import stats
    class1,class2=df[coly].unique()
    d={}
    for colx in colxs:
        _,d[colx]=getattr(stats,method)(df.loc[(df[coly]==class1),colx],
                                       df.loc[(df[coly]==class2),colx],
                                       alternative=alternative)
    return d    

def get_subset2metrics(df: pd.DataFrame,colvalue: str,colsubset: str,colindex: str,outstr=False,subset_control=None) -> dict:
    """Get subset-wise metrics.

    Args:
        df (DataFrame): input dataframe
        colvalue (str): column with values.
        colsubset (str): column with subsets.
        colindex (str): column with index.
        outstr (bool, optional): if output string. Defaults to False.
        subset_control (str, optional): control/reference subset. Defaults to None.

    Returns:
        dict: subset-wise metrics
    """
    if subset_control is None:
        subset_control=df[colsubset].unique().tolist()[-1]
    from scipy.stats import mannwhitneyu    
    df1=df.merge(df.loc[(df[colsubset]==subset_control),:],on=colindex, 
                 how='left',
                 suffixes=['',' reference'],
                )
    subset2metrics=df1.groupby(colsubset).apply(lambda df : mannwhitneyu(df[colvalue],df[f"{colvalue} reference"],
                                                     alternative='two-sided')).apply(pd.Series)[1].to_dict()
    if subset2metrics[subset_control]<0.9:
        logging.warning(f"pval for reference condition vs reference condition = {subset2metrics[subset_control]}. shouldn't be. check colindex")
    subset2metrics={k:subset2metrics[k] for k in subset2metrics if k!=subset_control}
    if outstr:
        from roux.viz.annot import pval2annot
        subset2metrics={k: pval2annot(subset2metrics[k],
                                      fmt='<',
                                      alternative='two-sided',
                                      linebreak=False) for k in subset2metrics}
    return subset2metrics

    
## for linear dfs
def get_demo_data():
    """Demo data to test the differences."""
    subsets=list('abcd')
    np.random.seed(88)
    df1=pd.concat({s:pd.Series([np.random.uniform(0,si+1) for _ in range((si+1)*100)]) for si,s in enumerate(subsets)},
             axis=0).reset_index(0).rename(columns={'level_0':'subset',0:'value'})
    df1['bool']=df1['value']>df1['value'].quantile(0.5)
    return df1

def compare_classes(x,y,method=None):
    """
    """
    if len(x)!=0 and len(y)!=0:# and (nunique(x+y)!=1):
        dplot=pd.crosstab(x,y)
        if (dplot.shape!=(2,2) and len(dplot)!=0) or method=='chi2':
            stat,pval,_,_=sc.stats.chi2_contingency(dplot)
            # stat_label='${\chi}^2$'
        else:
            stat,pval=sc.stats.fisher_exact(dplot)
            # stat_label='OR'
        return stat,pval
    else:
        return np.nan,np.nan
    
def compare_classes_many(
    df1: pd.DataFrame,
    cols_y: list,
    cols_x: list,
    ) -> pd.DataFrame:
    df0=pd.DataFrame(itertools.product(cols_y,cols_x,)).rename(columns={0:'colx',1:'coly'},errors='raise')
    # df0.head(1)
    # from roux.lib.stat.diff import compare_classes
    return (df0
            .join(df0.apply(lambda x: compare_classes(df1[x['colx']],
                                                      df1[x['coly']]),
                      axis=1)
            .apply(pd.Series)
            .rename(columns={0:'stat',1:'P'},errors='raise'))
           )
    
def get_pval(df: pd.DataFrame,
             colvalue='value',
             colsubset='subset',
             colvalue_bool=False,
             colindex=None,
             subsets=None,
            test=False,
            fun=None) -> tuple:
    """Get p-value.

    Args:
        df (DataFrame): input dataframe.
        colvalue (str, optional): column with values. Defaults to 'value'.
        colsubset (str, optional): column with subsets. Defaults to 'subset'.
        colvalue_bool (bool, optional): column with boolean values. Defaults to False.
        colindex (str, optional): column with the index. Defaults to None.
        subsets (list, optional): subset types. Defaults to None.
        test (bool, optional): test. Defaults to False.
        fun (function, optional): function. Defaults to None.

    Raises:
        ArgumentError: colvalue or colsubset not found in df.
        ValueError: need only 2 subsets.

    Returns:
        tuple: stat,p-value
    """
    if not ((colvalue in df) and (colsubset in df)):
        raise ArgumentError(f"colvalue or colsubset not found in df: {colvalue} or {colsubset}")
    if subsets is None:
        subsets=sorted(df[colsubset].unique())
    if len(subsets)!=2:
        raise ValueError('need only 2 subsets')
        return
    else:
        df=df.loc[df[colsubset].isin(subsets),:]
    if colvalue_bool and not df[colvalue].dtype==bool:        
        logging.warning(f"colvalue_bool {colvalue} is not bool")
        return
    if not colvalue_bool:      
#         try:
        x,y=df.loc[(df[colsubset]==subsets[0]),colvalue].tolist(),df.loc[(df[colsubset]==subsets[1]),colvalue].tolist()
        if len(x)!=0 and len(y)!=0 and (nunique(x+y)!=1):
            if fun is None:
                if test: logging.warning('mannwhitneyu used')               
                return sc.stats.mannwhitneyu(x,y,alternative='two-sided')
            else:
                logging.warning('custom function used')
                return fun(df.loc[(df[colsubset]==subsets[0]),colvalue],
                           df.loc[(df[colsubset]==subsets[1]),colvalue])
        else:
            #if empty list: RuntimeWarning: divide by zero encountered in double_scalars  z = (bigu - meanrank) / sd
            return np.nan,np.nan
    else:
        assert(not colindex is None)
        df1=df.pivot(index=colindex,columns=colsubset,values=colvalue)
        return compare_classes(df1[subsets[0]],df1[subsets[1]],method=None)
        # ct=pd.crosstab(df1[subsets[0]],df1[subsets[1]])
        # if ct.shape==(2,2):
        #     ct=ct.sort_index(axis=0,ascending=False).sort_index(axis=1,ascending=False)
        #     if test:
        #         print(ct)
        #     return sc.stats.fisher_exact(ct)
        # else:
        #     return np.nan,np.nan      
        
def get_stat(df1: pd.DataFrame,
              colsubset: str,
              colvalue: str,
              colindex: str,
              subsets=None,
              cols_subsets=['subset1', 'subset2'],
              df2=None,
              stats=[np.mean,np.median,np.var]+[len],
             coff_samples_min=None,
#               debug=False,
             verb=False,
             **kws,
             ) -> pd.DataFrame:
    """Get statistics.

    Args:
        df1 (DataFrame): input dataframe.
        colvalue (str, optional): column with values. Defaults to 'value'.
        colsubset (str, optional): column with subsets. Defaults to 'subset'.
        colindex (str, optional): column with the index. Defaults to None.
        subsets (list, optional): subset types. Defaults to None.
        cols_subsets (list, optional): columns with subsets. Defaults to ['subset1', 'subset2'].
        df2 (DataFrame, optional): second dataframe. Defaults to None.
        stats (list, optional): summary statistics. Defaults to [np.mean,np.median,np.var]+[len].
        coff_samples_min (int, optional): minimum sample size required. Defaults to None.
        verb (bool, optional): verbose. Defaults to False.
    
    Keyword Arguments:
        kws: parameters provided to `get_pval` function.

    Raises:
        ArgumentError: colvalue or colsubset not found in df.
        ValueError: len(subsets)<2

    Returns:
        DataFrame: output dataframe.
        
    TODOs:
        1. Rename to more specific `get_diff`, also other `get_stat*`/`get_pval*` functions.
    """
    if not ((colvalue in df1) and (colsubset in df1)):
        raise ArgumentError(f"colvalue or colsubset not found in df: {colvalue} or {colsubset}")
        return
    if subsets is None:
        subsets=sorted(df1[colsubset].unique())
    if len(subsets)<2:
        raise ValueError("len(subsets)<2")
        return
    colvalue_bool=df1[colvalue].dtype==bool
    if df2 is None:
        import itertools
        df2=pd.DataFrame([t for t in list(itertools.permutations(subsets,2))] if len(subsets)>2 else [subsets])
        df2.columns=cols_subsets
    df2=df2.groupby(cols_subsets).apply(lambda df: get_pval(df1,colvalue=colvalue,
                                                            colsubset=colsubset,
                                                            colindex=colindex,
                                                            subsets=df.name,
                                                            colvalue_bool=colvalue_bool,
                                                            **kws,
                                                           )).apply(pd.Series)
    df2=df2.rename(columns={0:f"stat ({'MWU' if not colvalue_bool else 'FE'} test)",
                            1:f"P ({'MWU' if not colvalue_bool else 'FE'} test)",
                            }).reset_index()
    from roux.lib.dfs import merge_paired
    from roux.lib.str import get_prefix,get_suffix
    colsubset_=get_prefix(*cols_subsets,common=True, clean=True)
    df_=(df1
        .groupby([colsubset])[colvalue]
        .agg(stats if not colvalue_bool else [sum,len])
        .reset_index()
             # TODOs rename to subset1 subset2
        .rename(columns={colsubset:colsubset_},errors='raise'))        
    df3=merge_paired(
        df1=df2,
        df2=df_,
        left_ons=cols_subsets,
        right_on=colsubset_,
        common=[],
        right_ons_common=[],
        how='inner',
        validates=['m:1', 'm:1'],
        suffixes=get_suffix(*cols_subsets,common=False, clean=False),
        test=False,
        verb=False,
        # **kws,
    )
    df3=df3.rename(columns={f"{c}{i}":f"{c} {colsubset_}{i}" for c in df_ for i in [1,2] if c!=colsubset_},errors='raise')
    ## minimum samples
    if not coff_samples_min is None:
        # logging.info("coff_samples_min applied")
        df3.loc[((df3[f'len {cols_subsets[0]}']<coff_samples_min) \
                 | (df3[f'len {cols_subsets[1]}']<coff_samples_min)),
                df3.filter(like="P (").columns.tolist()]=np.nan
    return df3

def get_stats(df1: pd.DataFrame,
              colsubset: str,
              cols_value: list,
              colindex: str,
              subsets=None,
              df2=None,
              cols_subsets=['subset1', 'subset2'],
              stats=[np.mean,np.median,np.var,len],
              axis=0, # concat 
              test=False,
              **kws) -> pd.DataFrame:
    """Get statistics by iterating over columns wuth values.

    Args:
        df1 (DataFrame): input dataframe.
        colsubset (str, optional): column with subsets.
        cols_value (list): list of columns with values.
        colindex (str, optional): column with the index.
        subsets (list, optional): subset types. Defaults to None.
        df2 (DataFrame, optional): second dataframe, e.g. `pd.DataFrame({"subset1":['test'],"subset2":['reference']})`. Defaults to None.
        cols_subsets (list, optional): columns with subsets. Defaults to ['subset1', 'subset2'].
        stats (list, optional): summary statistics. Defaults to [np.mean,np.median,np.var]+[len].
        axis (int, optional): 1 if different tests else use 0. Defaults to 0.
    
    Keyword Arguments:
        kws: parameters provided to `get_pval` function.

    Raises:
        ArgumentError: colvalue or colsubset not found in df.
        ValueError: len(subsets)<2

    Returns:
        DataFrame: output dataframe.

    TODOs:
        1. No column prefix if `len(cols_value)==1`.

    """
    dn2df={}
    if subsets is None:
        subsets=sorted(df1[colsubset].unique())
    from roux.stat.diff import get_stat ## remove?
    for colvalue in cols_value:
        df1_=df1.dropna(subset=[colsubset,colvalue])
        if len(df1_[colsubset].unique())>1:
            dn2df[colvalue]=get_stat(df1_,
                          colsubset=colsubset,
                          colvalue=colvalue,
                          colindex=colindex,
                          subsets=subsets,
                          cols_subsets=cols_subsets,
                          df2=df2,
                          stats=stats,
                          **kws,
                         ).set_index(cols_subsets)
        else:
            if test:
                logging.warning(f"not processed: {colvalue}; probably because of dropna")
    if len(dn2df.keys())==0:
        return 
    import pandas as pd # remove?
    df3=pd.concat(dn2df,
                  ignore_index=False,
                  axis=axis,
                  verify_integrity=True,
                  names=None if axis==1 else ['variable'],
                 )
    if axis==1:
        df3=df3.reset_index().rd.flatten_columns()
    return df3

def get_q(ds1,col=None,verb=True,test_coff=0.1):
    if not col is None:
        df1=ds1.copy()
        ds1=ds1[col]
    ds2=ds1.dropna()
    from statsmodels.stats.multitest import fdrcorrection
    ds3=fdrcorrection(pvals=ds2, alpha=0.05, method='indep', is_sorted=False)[1]
    ds4=ds1.map(pd.DataFrame({'P':ds2,'Q':ds3}).drop_duplicates().rd.to_dict(['P','Q']))
    if verb:
        from roux.viz.annot import perc_label        
        logging.info(f"significant at Q<{test_coff}: {perc_label(ds4<test_coff)}")
    if col is None:
        return ds4
    else:
        df1['Q']=ds4
        return df1

def get_significant_changes(df1: pd.DataFrame,
                            coff_p=0.025,
                            coff_q=0.1,
                            alpha=None,
                            changeby="mean",
                            # fdr=True,
                            value_aggs=['mean','median'],
                           ) -> pd.DataFrame:
    """Get significant changes.

    Args:
        df1 (DataFrame): input dataframe.
        coff_p (float, optional): cutoff on p-value. Defaults to 0.025.
        coff_q (float, optional): cutoff on q-value. Defaults to 0.1.
        alpha (float, optional): alias for `coff_p`. Defaults to None.
        changeby (str, optional): "" if check for change by both mean and median. Defaults to "".
        value_aggs (list, optional): values to aggregate. Defaults to ['mean','median'].

    Returns:
        DataFrame: output dataframe.
    """
    if coff_p is None and not alpha is None:
        coff_p=alpha
    if df1.filter(regex='|'.join([f"{s} subset(1|2)" for s in value_aggs])).shape[1]:
        for s in value_aggs:
            df1[f'difference between {s} (subset1-subset2)']=df1[f'{s} subset1']-df1[f'{s} subset2']
        ## call change if both mean and median are changed
        # df1.loc[((df1.filter(like=f'difference between {changeby}')>0).T.sum()==2),'change']='increase'
        # df1.loc[((df1.filter(like=f'difference between {changeby}')<0).T.sum()==2),'change']='decrease'
        info(changeby)
        df1.loc[(df1[f'difference between {changeby} (subset1-subset2)']>0),'change']='increase'
        df1.loc[(df1[f'difference between {changeby} (subset1-subset2)']<0),'change']='decrease'
        df1['change']=df1['change'].fillna('ns')
    from statsmodels.stats.multitest import multipletests
    for test in ['MWU','FE']:
        if not f'P ({test} test)' in df1:
            continue
        # without fdr
        df1[f'change is significant, P ({test} test) < {coff_p}']=df1[f'P ({test} test)']<coff_p
        if not coff_q is None:
            df1[f'Q ({test} test)']=get_q(df1[f'P ({test} test)'])
            # df1[f'change is significant, Q ({test} test) < {coff_q}']=df1[f'Q ({test} test)']<coff_q
        #     info(f"corrected alpha alphacSidak={alphacSidak},alphacBonf={alphacBonf}")
        # if test!='FE':
            df1[f"significant change, Q ({test} test) < {coff_q}"]=df1.apply(lambda x: x['change'] if x[f'Q ({test} test)']<coff_q else 'ns',axis=1)
            # df1.loc[df1[f'change is significant, Q ({test} test) < {coff_q}'],f"significant change, Q ({test} test) < {coff_q}"]=df1.loc[df1[f"change is significant, Q ({test} test) < {coff_q}"],'change']
            # df1[f"significant change, Q ({test} test) < {coff_q}"]=df1[f"significant change, Q ({test} test) < {coff_q}"].fillna('ns')
    return df1

def apply_get_significant_changes(df1: pd.DataFrame,cols_value: list,
                                    cols_groupby: list, # e.g. genes id
                                    cols_grouped: list, # e.g. tissue
                                    fast=False,
                                    **kws,
                                    ) -> pd.DataFrame:
    """Apply on dataframe to get significant changes.

    Args:
        df1 (DataFrame): input dataframe.
        cols_value (list): columns with values.
        cols_groupby (list): columns with groups.

    Returns:
        DataFrame: output dataframe.
    """
    d1={}
    from tqdm import tqdm
    for c in tqdm(cols_value):
        df1_=df1.set_index(cols_groupby).filter(regex=f"^{c} .*").filter(regex="^(?!("+' |'.join([s for s in cols_value if c!=s])+")).*")
        df1_=df1_.rd.renameby_replace({f'{c} ':''}).reset_index()
        d1[c]=getattr(df1_.groupby(cols_groupby),'apply' if not fast else 'parallel_apply')(lambda df: get_significant_changes(df,**kws))
    d1={k:d1[k].set_index(cols_groupby) for k in d1}
    ## to attach the info
    d1['grouped']=df1.set_index(cols_groupby).loc[:,cols_grouped]
    df2=pd.concat(d1,
                  ignore_index=False,
                  axis=1,
                  verify_integrity=True,
                 )    
    df2=df2.rd.flatten_columns().reset_index()
    assert(not df2.columns.duplicated().any())
    return df2

def get_stats_groupby(df1: pd.DataFrame,cols: list,
                      coff_p: float=0.05,
                      coff_q: float=0.1,
                      alpha=None,
                      fast=False,
                      **kws) -> pd.DataFrame:
    """Iterate over groups, to get the differences.

    Args:
        df1 (DataFrame): input dataframe.
        cols (list): columns to interate over.
        coff_p (float, optional): cutoff on p-value. Defaults to 0.025.
        coff_q (float, optional): cutoff on q-value. Defaults to 0.1.
        alpha (float, optional): alias for `coff_p`. Defaults to None.
        fast (bool, optional): parallel processing. Defaults to False.

    Returns:
        DataFrame: output dataframe.
    """
    df2=getattr(df1.groupby(cols),f"{'progress' if not fast else 'parallel'}_apply")(lambda df: get_stats(df1=df,**kws)).reset_index().rd.clean()
    return get_significant_changes(df1=df2,alpha=alpha,coff_p=coff_p,coff_q=coff_q,)

def get_diff(
    df1,
    cols_x,
    cols_y,
    cols_index,
    test=True,
    **kws
    )-> pd.DataFrame:
    """
    Wrapper around the `get_stats_groupby`
    
    Keyword parameters:
          cols=['variable x','variable y'],
          coff_p=0.05,
          coff_q=0.01,
          colindex=['id'],    
    """
    ## filter the significant 
    d_={}
    for colx in cols_x:
        assert df1[colx].nunique()==2,colx
        d_[colx]=(df1
        .melt(id_vars=cols_index+[colx],
                 value_vars=cols_y,
                 var_name='variable y',
                 value_name='value y')
        .rename(columns={colx:'value x'},errors='raise')
        )
    df2=pd.concat(d_,names=['variable x']).reset_index().rd.clean().log.dropna()
    if test:
        info(df2.iloc[0,:])
    df3=get_stats_groupby(df1=df2,
                      colsubset='value x',
                      cols_value= ['value y'],
                          colindex=cols_index,
                          **kws,
                     )
    return df3.loc[(df3['P (MWU test)']<0.05),:].sort_values('P (MWU test)')

def binby_pvalue_coffs(df1: pd.DataFrame,
                    coffs=[0.01,0.05,0.25],
                    color=False,
                    testn='MWU test, FDR corrected',
                    colindex='genes id',
                    colgroup='tissue',
                    preffix='',
                    colns=None, # plot as ns, not counted
                    palette=None,#['#f55f5f','#ababab','#046C9A',],
                      ) -> tuple:
    """Bin data by pvalue cutoffs.

    Args:
        df1 (DataFrame): input dataframe.
        coffs (list, optional): cut-offs. Defaults to [0.01,0.05,0.25].
        color (bool, optional): color asignment. Defaults to False.
        testn (str, optional): test number. Defaults to 'MWU test, FDR corrected'.
        colindex (str, optional): column with index. Defaults to 'genes id'.
        colgroup (str, optional): column with the groups. Defaults to 'tissue'.
        preffix (str, optional): prefix. Defaults to ''.
        colns (_type_, optional): columns number. Defaults to None.
        notcountedpalette (_type_, optional): _description_. Defaults to None.

    Returns:
        tuple: output.
    
    Notes:
        1. To be deprecated in the favor of the functions used for enrichment analysis for example. 
    """
    assert(len(df1)!=0)
    if palette is None:
        from roux.viz.colors import get_colors_default
        palette=get_colors_default()[:3]
    assert(len( palette)==3)
    coffs=np.array(sorted(coffs))
    # df1[f'{preffix}P (MWU test, FDR corrected) bin']=pd.cut(x=df1[f'{preffix}P (MWU test, FDR corrected)'],
    #       bins=[0]+coffs+[1],
    #        labels=coffs+[1],
    #        right=False,
    #       ).fillna(1)
    from roux.viz.colors import saturate_color
    d1={}
    for i,coff in enumerate(coffs[::-1]):
        col=f"{preffix}significant change, P ({testn}) < {coff}"
        df1[col]=df1.apply(lambda x: 'increase' if ((x[f'{preffix}P ({testn})']<coff) \
                                                    and (x[f'{preffix}difference between mean (subset1-subset2)']>0))\
                                else 'decrease' if ((x[f'{preffix}P ({testn})']<coff) \
                                                    and (x[f'{preffix}difference between mean (subset1-subset2)']<0))\
                                else 'ns', axis=1)
        if color:
            if i==0:
                df1.loc[(df1[col]=='ns'),'c']=palette[1]
            saturate=1-((len(coffs)-(i+1))/len(coffs))
            d2={}
            d2['increase']=saturate_color(palette[0],saturate)
            d2['decrease']=saturate_color(palette[2],saturate)
            d1[coff]=d2
            df1['c']=df1.apply(lambda x: d2[x[col]] if x[col] in d2 else x['c'],axis=1)
            assert(df1['c'].isnull().sum()==0)
    if color:
        import itertools
        from roux.stat.transform import rescale,log_pval
        d3={}
        for i,(k,coff) in enumerate(list(itertools.product(['increase','decrease'],coffs))):
            col=f"{preffix}significant change, P ({testn}) < {coff}"
            d4={}
            d4['y alpha']=rescale(1-(list(coffs).index(coff))/len(coffs),[0,1],[0.5,1])
            d4['y']=log_pval(coff)
            d4['y text']=f" P < {coff}"
            d4['x']=df1.loc[(df1[col]==k),f'{preffix}difference between mean (subset1-subset2)'].min() if k=='increase' \
                                        else df1.loc[(df1[col]==k),f'{preffix}difference between mean (subset1-subset2)'].max()
            d4['change']=k
            d4['text']=f"{df1.loc[(df1[col]==k),colindex].nunique()}/{df1.loc[(df1[col]==k),colgroup].nunique()}"
            d4['color']=d1[coff][k]
            d3[i]=d4
        df2=pd.DataFrame(d3).T
    if not colns is None:
        df1.loc[df1[colns],'c']=palette[1]
#     info(df1.shape,df1.shape)
    return df1,df2

# from roux.viz.diff import plot_stats_diff
## confounding effects
def get_stats_regression(df_: pd.DataFrame,
                        d0={},
                        variable=None,
                        covariates=None,
                        converged_only=False,
                         out='df',
                        verb=False,
                        test=False,
                        **kws,
                        ) -> pd.DataFrame:
    """Get stats from regression models.

    Args:
        df_ (DataFrame): input dataframe.
        d0 (dict, optional): model name to base equation e.g. 'y ~ x'. Defaults to {}.
        variable (_type_, optional): to get params of e.g. 'C(variable)[T.True]'. Defaults to None.
        covariates (_type_, optional): variables. Defaults to None.
        converged_only (bool, optional): get the stats from the converged models only. Defaults to False.
        out (str, optional): output format. Defaults to 'df'.
        verb (bool, optional): verbose. Defaults to False.
        test (bool, optional): test. Defaults to False.

    Returns:
        DataFrame: output.
    """
    if test and hasattr(df_,'name'):
        info(df_.name)
    def to_df(res):
        if isinstance(res.summary().tables[1],pd.DataFrame):
            df1=res.summary().tables[1]
        elif hasattr(res.summary().tables[1],'as_html'):
            df1=pd.read_html(res.summary().tables[1].as_html(), header=0, index_col=0)[0]
        else:
            logging.error('dataframe not found')
            return
        df1.columns.name='stat'
        df1.index.name='variable'
        return df1.melt(ignore_index=False).reset_index()
    def get_stats(res,variable):
        return pd.Series([res.pvalues[variable], 
                          res.params[variable], 
                        ], 
                        index=['P', 'coefficient',
                        ]).to_frame('value')
    if not (verb or test):
        import warnings
        from statsmodels.tools.sm_exceptions import ConvergenceWarning
        warnings.simplefilter('ignore', ConvergenceWarning)
        warnings.simplefilter('ignore', RuntimeWarning)
        warnings.simplefilter('ignore', UserWarning)
    import statsmodels.formula.api as smf
    from statsmodels.tools.sm_exceptions import PerfectSeparationError
    from numpy.linalg import LinAlgError
    if not covariates is None:
        #formats
        d1=df_.dtypes.to_dict()
        formula_covariates=' + '+' + '.join([k if ((d1[k]==int) or (d1[k]==float)) else f"C({k})" for k in covariates if k in d1])
    else:
        formula_covariates=''
    # info(formula_covariates)
    d1={}
    for k,formula_base in d0.items():
        if isinstance(k,str):
            model=getattr(smf,k)
        elif isinstance(k,object) and hasattr(k,'from_formula'):
            model=k.from_formula
        elif isinstance(k,object):
            model=k
        else:
            logging.error(model)
            return
        # print(str(model))
        formula=formula_base+formula_covariates
        if test: info(formula)
        modeln=str(model).split('.')[-1].split("'")[0]
        if not 'groups' in kws:
            try:
                d1[modeln]=model(data=df_,formula=formula).fit(disp=False)
            except (PerfectSeparationError,LinAlgError) as e:
                if verb or test: logging.error('PerfectSeparationError/LinAlgError')
        else:
            try:
                d1[modeln]=model(data=df_,formula=formula,
                                groups=df_[kws["groups"]],
                                ).fit(disp=False)
            except (PerfectSeparationError,LinAlgError) as e:
                if verb or test: logging.error('PerfectSeparationError/LinAlgError')
    if out=='model':
        return d1
    elif out=='df':
        d1={k:to_df(v) for k,v in d1.items() if ((hasattr(v,'converged') and (v.converged)) or (not converged_only))}
        if len(d1)!=0:
            if not variable is None:
                return pd.concat({k:get_stats(v,variable=variable) for k,v in d1.items()},
                                 axis=0,names=['model type','variable']).reset_index()
            else:
                return pd.concat(d1,axis=0,names=['model type']).reset_index(0)
    
def filter_regressions(df1: pd.DataFrame,
                       variable: str,
                       colindex: str,
                       coff_q : float=0.1,
                       by_covariates: bool=True,
                       coff_p_covariates: float=0.05,
                       test: bool=False) -> pd.DataFrame:
    """Filter regression statistics.

    Args:
        df1 (DataFrame): input dataframe.
        variable (str): variable name to filter by.
        colindex (str): columns with index.
        coff_q (float, optional): cut-off on the q-value. Defaults to 0.1.
        by_covariates (bool, optional): filter by these covaliates. Defaults to True.
        coff_p_covariates (float, optional): cut-off on the p-value for the covariates. Defaults to 0.05.
        test (bool, optional): test. Defaults to False.

    Raises:
        ValueError: pval.

    Returns:
        DataFrame: output.
    """
    pval='P>|t|' if 'P>|t|' in df1['stat'].tolist() else 'P>|z|' if 'P>|z|' in df1['stat'].tolist() else None
    if pval is None:
        raise ValueError(pval)
    df1['stat']=df1['stat'].apply(lambda x: 'P' if x==pval else x )
    df2=df1.loc[((df1['variable']==variable) & (df1['stat'].isin(['coef','P']))),:]
#     print(df1['stat'].unique())
    df3=df2.pivot(index=colindex,columns='stat',values='value').reset_index()
#     print(df3.columns)
    df3=df3.rename(columns={'coef':'score'},
                   errors='raise')
    df3=df3.log.dropna(subset=['P'])
    if test:
        df3['P'].hist()
    from statsmodels.stats.multitest import fdrcorrection
    df3['Q']=fdrcorrection(pvals=df3['P'], alpha=0.05, method='indep', is_sorted=False)[1]
    if test:
        df3['Q'].hist()
    info(sum(df3['P']<coff_q))
    info(sum(df3['Q']<coff_q))
    df3=df3.loc[(df3['Q']<coff_q),:]
    info(df3[colindex].nunique())
    if by_covariates:
        ## #2
        ## all covariates (potential confounding effects) are non-sinificant
        ## non-standardised regression 'P>|t|' or standardised 'P>|z|'
        df5=df1.loc[(
                (df1['variable']!='Intercept') \
              & (df1['variable']!=variable) \
              & (df1['stat']=='P')
                ),:]
        df6=df5.groupby(colindex).filter(lambda df: (df['value']>=coff_p_covariates).all()).loc[:,[colindex]]
        df3=df3.log().loc[df3[colindex].isin(df6[colindex]),:].log()
    else:
        logging.warning("not filtered by_covariates")
    info(df3[colindex].nunique())
    return df3
    
import statsmodels.api as sm
import statsmodels.formula.api as smf

def get_model_summary(model: object) -> pd.DataFrame:
    """Get model summary.

    Args:
        model (object): model.

    Returns:
        pd.DataFrame: output.
    """
    df_=model.summary().tables[0]
    return df_.loc[:,[0,1]].append(df_.loc[:,[2,3]].rename(columns={2:0,3:1})).rename(columns={0:'index',1:'value'}).append(dmap2lin(model.summary().tables[1]),sort=True)

def run_lr_test(data: pd.DataFrame,
                formula: str,
                covariate: str,
                col_group: str,
                params_model: dict ={'reml':False}
                ) -> tuple:
    """Run LR test.

    Args:
        data (pd.DataFrame): input data.
        formula (str): formula.
        covariate (str): covariate.
        col_group (str): column with the group.
        params_model (dict, optional): parameters of the model. Defaults to {'reml':False}.

    Returns:
        tuple: output tupe (stat, pval,dres).
    """

    sc.stats.chisqprob = lambda chisq, df: sc.stats.chi2.sf(chisq, df)
    def get_lrtest(llmin, llmax):
        stat = 2 * (llmax - llmin)
        pval = sc.stats.chisqprob(stat, 1)
        return stat, pval        
    data=data.dropna()
    # without covariate
    model = smf.mixedlm(formula, data,groups=data[col_group])
    modelf = model.fit(**params_model)
    llf = modelf.llf

    # with covariate
    model_covariate = smf.mixedlm(f"{formula}+ {covariate}", data,groups=data[col_group])
    modelf_covariate = model_covariate.fit(**params_model)
    llf_covariate = modelf_covariate.llf

    # compare
    stat, pval = get_lrtest(llf, llf_covariate)
    print(f'stat {stat:.2e} pval {pval:.2e}')
    
    # results
    dres=delunnamedcol(pd.concat({False:get_model_summary(modelf),
    True:get_model_summary(modelf_covariate)},axis=0,names=['covariate included','Unnamed']).reset_index())
    return stat, pval,dres

def plot_residuals_versus_fitted(model: object) -> plt.Axes:
    """plot Residuals Versus Fitted (RVF).

    Args:
        model (object): model.

    Returns:
        plt.Axes: output.
    """
    fig = plt.figure(figsize = (5, 3))
    ax = sns.scatterplot(y = model.resid, x = model.fittedvalues,alpha=0.2)
    ax.set_xlabel("fitted")
    ax.set_ylabel("residuals")
    l = sm.stats.diagnostic.het_white(model.resid, model.model.exog)
    ax.set_title("LM test "+pval2annot(l[1],alpha=0.05,fmt='<',linebreak=False)+", FE test "+pval2annot(l[3],alpha=0.05,fmt='<',linebreak=False))    
    return ax

def plot_residuals_versus_groups(model: object) -> plt.Axes:
    """plot Residuals Versus groups.

    Args:
        model (object): model.

    Returns:
        plt.Axes: output.
    """
    fig = plt.figure(figsize = (5, 3))
    ax = sns.pointplot(x = model.model.groups, 
                       y = model.resid,
                      ci='sd',
                      join=False)
    ax.set_ylabel("residuals")
    ax.set_xlabel("groups")
    return ax

def plot_model_sanity(model: object):
    """Plot sanity stats.

    Args:
        model (object): model.
    """
    from roux.viz.scatter import plot_qq 
    from roux.viz.dist import plot_normal 
    plot_normal(x=model.resid)
    plot_qq(x=model.resid)
    plot_residuals_versus_fitted(model)
    plot_residuals_versus_groups(model)



import numpy as np

def get_intersection_locations(y1: np.array,y2: np.array,test: bool=False,x: np.array=None) -> list: 
    """Get co-ordinates of the intersection (x[idx]).

    Args:
        y1 (np.array): vector.
        y2 (np.array): vector.
        test (bool, optional): test mode. Defaults to False.
        x (np.array, optional): vector. Defaults to None.

    Returns:
        list: output.
    """
    idxs=np.argwhere(np.diff(np.sign(y1 - y2))).flatten()
    if test:
        x=range(len(y1)) if x is None else x
        plt.figure(figsize=[2.5,2.5])
        ax=plt.subplot()
        ax.plot(x,y1,color='r',label='line1',alpha=0.5)
        ax.plot(x,y2,color='b',label='line2',alpha=0.5)
        _=[ax.axvline(x[i],color='k') for i in idxs]
        _=[ax.text(x[i],ax.get_ylim()[1],f"{x[i]:1.1f}",ha='center',va='bottom') for i in idxs]
        ax.legend(bbox_to_anchor=[1,1])
        ax.set(xlabel='x',ylabel='density')
    return idxs

# def get_intersection_of_gaussians(m1, s1, m2, s2):
#     # coefficients of quadratic equation ax^2 + bx + c = 0
#     a = (s1**2.0) - (s2**2.0)
#     b = 2 * (m1 * s2**2.0 - m2 * s1**2.0)
#     c = m2**2.0 * s1**2.0 - m1**2.0 * s2**2.0 - 2 * s1**2.0 * s2**2.0 * np.log(s1/s2)
#     x1 = (-b + np.sqrt(b**2.0 - 4.0 * a * c)) / (2.0 * a)
#     x2 = (-b - np.sqrt(b**2.0 - 4.0 * a * c)) / (2.0 * a)
#     return x1, x2

# def get_intersection_of_gaussians(m1,m2,std1,std2):
#     a = 1/(2*std1**2) - 1/(2*std2**2)
#     b = m2/(std2**2) - m1/(std1**2)
#     c = m1**2 /(2*std1**2) - m2**2 / (2*std2**2) - np.log(std2/std1)
#     return np.roots([a,b,c])


"""Dictionary processing"""
from roux.lib.set import *
import yaml
import json
from roux.lib.str import dict2str,str2dict

def head_dict(d, lines=5):
    return dict(itertools.islice(d.items(), lines))

def sort_dict(d1,by=1,ascending=True):
    """Sort dictionary by values.
    
    Parameters:
        d1 (dict): input dictionary.
        by (int): index of the value among the values.
        ascending (bool): ascending order.
        
    Returns:
        d1 (dict): output dictionary.
    """
    return dict(sorted(d1.items(), key=lambda item: item[1],reverse=not ascending))

def merge_dicts(l):
    """Merge dictionaries.
    
    Parameters:
        l (list): list containing the dictionaries.
        
    Returns:
        d (dict): output dictionary.
    """    
    from collections import ChainMap
    return dict(ChainMap(*l))

def merge_dict_values(l,test=False):
    """Merge dictionary values.
    
    Parameters:
        l (list): list containing the dictionaries.
        test (bool): verbose.
        
    Returns:
        d (dict): output dictionary.
    """
    for di,d_ in enumerate(l):
        if di==0:
            d=d_
        else:
            d={k:d[k]+d_[k] for k in d}
        if test:
            print(','.join([str(len(d[k])) for k in d]))
    return d    

def read_yaml(p):
    """Read `.yaml` file.
    
    Parameters:
        p (str): path.
        
    Returns:
        d (dict): output dictionary.
    """
    with open(p,'r') as f:
        return yaml.safe_load(f)
def to_yaml(d,p,**kws): 
    """Save `.yaml` file.
    
    Parameters:
        d (dict): input dictionary.
        p (str): path.
        
    Keyword Arguments:
        kws (d): parameters provided to `yaml.safe_dump`.
        
    Returns:
        p (str): path.
    """
    with open(p,'w') as f:
        yaml.safe_dump(d,f,**kws)
    return p        
def read_json(path_to_file,encoding=None):    
    """Read `.json` file.
    
    Parameters:
        p (str): path.
        
    Returns:
        d (dict): output dictionary.
    """
    with open(path_to_file,encoding=encoding) as p:
        return json.load(p)
def to_json(data,p):
    """Save `.json` file.
    
    Parameters:
        d (dict): input dictionary.
        p (str): path.
                
    Returns:
        p (str): path.
    """
    with open(p, 'w') as outfile:
        json.dump(data, outfile)
    return p

def read_pickle(p):
    """Read `.pickle` file.
    
    Parameters:
        p (str): path.
        
    Returns:
        d (dict): output dictionary.
    """    
    import pickle
    return pickle.load(open(p,
               'rb'))
def is_dict(p):
    return p.endswith(('.yml','.yaml','.json','.joblib','.pickle'))
    
def read_dict(p,fmt='',**kws):
    """Read dictionary file.
    
    Parameters:
        p (str): path.
        fmt (str): format of the file.
        
    Keyword Arguments:
        kws (d): parameters provided to reader function.
        
    Returns:
        d (dict): output dictionary.
    """    
    if '*' in p:
        from roux.lib.io import basenamenoext
        from glob import glob
        return {basenamenoext(p):read_dict(p) for p in glob(p)}
    if p.endswith('.yml') or p.endswith('.yaml') or fmt=='yml' or fmt=='yaml':
        return read_yaml(p,**kws)
    elif p.endswith('.json') or fmt=='json':
        return read_json(p,**kws)
    elif p.startswith('https'):
        from urllib.request import urlopen
        try:
            return json.load(urlopen(p))
        except:
            print(logging.error(p))
#         return read_json(p,**kws)    
    elif p.endswith('.pickle'):
        return read_pickle(p,**kws)
    elif p.endswith('.joblib'):
        import joblib
        return joblib.load(p,**kws)
    else:
        logging.error(f'supported extensions: .yml .yaml .json .pickle .joblib')
        
def to_dict(d,p,**kws):
    """Save dictionary file.
    
    Parameters:
        d (dict): input dictionary.
        p (str): path.
                
    Keyword Arguments:
        kws (d): parameters provided to export function.
        
    Returns:
        p (str): path.
    """
    from roux.lib.sys import makedirs
    if not 'My Drive' in p:
        p=p.replace(' ','_')
    else:
        logging.warning('probably working on google drive; space/s left in the path.')
    makedirs(p)
    if p.endswith('.yml') or p.endswith('.yaml'):
        return to_yaml(d,p,**kws)
    elif p.endswith('.json'):
        return to_json(d,p,**kws)
    elif p.endswith('.pickle'):
        import pickle
        return pickle.dump(d, open(p, 'wb'),**kws)
    elif p.endswith('.joblib'):
        import joblib
        return joblib.dump(d, p,**kws)     
    else:
        raise ValueError(f'supported extensions: .yml .yaml .json .pickle .joblib')        
        
def flip_dict(d):
    """switch values with keys and vice versa.
    
    Parameters:
        d (dict): input dictionary.
        
    Returns:
        d (dict): output dictionary.
    """
    if all([not isinstance(s,list) for s in d.values()]):
        if len(np.unique(d.keys()))!=len(np.unique(d.values())):
            logging.error('values should be unique to flip the dict')
            return
        else:
            return {d[k]:k for k in d}
    else:
        if not get_offdiagonal_values(intersections(d)).sum().sum()==0:
            raise ValueError('dict values should be mutually exclusive') 
        d_={}
        for k in d:
            for v in d[k]:
                d_[v]=k
        return d_



from pathlib import Path
from roux.lib.sys import logging,exists,dirname,basename,makedirs,basenamenoext,info,abspath
from roux.lib.io import read_ps
import pandas as pd
import numpy as np

def get_scripts(
    ps: list,
    notebook_prefix: str='\d{2}',
    notebook_suffix: str='_v\d{2}',
    test: bool=False, 
    fast: bool=True,
    cores: int=6,
    force: bool=False,
    tab: str='    ',
    **kws,
    ) -> pd.DataFrame:
    """Get scripts.

    Args:
        ps (list): paths.
        notebook_prefix (str, optional): prefix of the notebook file to be considered as a "task".
        notebook_suffix (str, optional): suffix of the notebook file to be considered as a "task".
        test (bool, optional): test mode. Defaults to False.
        fast (bool, optional): parallel processing. Defaults to True.
        cores (int, optional): cores to use. Defaults to 6.
        force (bool, optional): overwrite the outputs. Defaults to False.
        tab (str, optional): tab in spaces. Defaults to '    '.

    Returns:
        pd.DataFrame: output table.
    """
    d1={basenamenoext(p):p for p in ps}
    # print(d1)
    # d1={p:v for p,v in d1.items() if p.startswith(notebook_prefix) and p.endswith(notebook_suffix)}
    import re
    d1={k:v for k,v in d1.items() if re.match(notebook_prefix+'.*'+notebook_suffix,k)}    
    # print(d1)
    if test: info(d1)
    assert len(d1)!=0, f'no notebooks with the {notebook_prefix} prefix and {notebook_suffix} suffix found'
    if notebook_suffix!='':
        ## remove suffix
        # d1={p.split(notebook_suffix)[0]:v for p,v in d1.items()}
        d1={re.split(notebook_suffix, p)[0].rstrip('_'):v for p,v in d1.items()}
    # print(d1.keys())
    # brk
    df1=pd.DataFrame(pd.Series(d1,name='notebook path'))
    df1.index.name='task name'
    df1=df1.sort_index().reset_index()
    if not df1['task name'].apply(lambda x: x[:1].isdigit()).all():
        logging.warning('notebooks are not numbered')
    # df1=df1.loc[(df1['notebook path'].apply(lambda x: (notebook_suffix in x))),:]
    # print(df1.shape)
    from roux.workflow.function import to_task
    if not fast or df1['notebook path'].nunique()<5:
        # df2=df1['notebook path'].apply(lambda x: print(x,force,not fast,notebook_suffix,kws))
        df2=df1.apply(lambda x: to_task(notebookp=x['notebook path'],
                                        task=x['task name'],
                                         force=force,verbose=not fast,
                                         notebook_suffix=notebook_suffix,
                                         **kws),
                                         axis=1)
    else:
        from roux.lib.df import get_name
        from pandarallel import pandarallel
        pandarallel.initialize(nb_workers=cores,progress_bar=True)
        df2=df1.groupby(['notebook path','task name']).parallel_apply(lambda x: to_task(
                                                                            notebookp=get_name(x,['notebook path'])[0],
                                                                            task=get_name(x,['task name'])[0],
                                                                            force=force,verbose=not fast,
                                                                            notebook_suffix=notebook_suffix,
                                                                            **kws))
    assert len(df2)!=0, 'no notebooks found'
    df2=df2.apply(pd.Series).rename(columns={0:'rule code',1:'output paths'},errors='raise')
    if df2.index.name is None:
        df2.index.name='notebook path'
    # print(df2)
    df2=df2.reset_index(0).dropna(subset=['notebook path','rule code'])
    if df1['notebook path'].nunique()==1:
        # print(df2.index)
        # print(df2.shape)
        # print(df2)
        df2['notebook path']=df1['notebook path'].tolist()[0]
    from roux.lib.set import flatten
    # if test:print(d1)
    df2['output paths']=df2['output paths'].apply(lambda x: f",\n{tab}{tab}".join(flatten(x)) if isinstance(x,list) else '')
    df2=df2.reset_index(drop=True)
    assert len(df2)!=0, 'no functions found'
    return df2

def to_scripts(
    # packagen: str,
    packagep: str,
    notebooksdp: str,
    validate: bool=False,
    ps: list=None,
    notebook_prefix: str='\d{2}',
    notebook_suffix: str='_v\d{2}',
    scripts: bool=True,
    workflow: bool=True,
    sep_step: str='## step',
    todos: bool=False,
    git: bool=True,
    clean: bool=False,
    test: bool=False,    
    force: bool=True,
    tab: str='    ',
    **kws,
    ):
    """To scripts.

    Args:
        # packagen (str): package name.
        packagep (str): path to the package.
        notebooksdp (str, optional): path to the notebooks. Defaults to None.
        validate (bool, optional): validate if functions are formatted correctly. Defaults to False.
        ps (list, optional): paths. Defaults to None.
        notebook_prefix (str, optional): prefix of the notebook file to be considered as a "task".
        notebook_suffix (str, optional): suffix of the notebook file to be considered as a "task".
        scripts (bool, optional): make scripts. Defaults to True.
        workflow (bool, optional): make workflow file. Defaults to True.
        sep_step (str, optional): separator marking the start of a step. Defaults to "## step".
        todos (bool, optional): show todos. Defaults to False.
        git (bool, optional): save version. Defaults to True.
        clean (bool, optional): clean temporary files. Defaults to False.
        test (bool, optional): test mode. Defaults to False.
        force (bool, optional): overwrite outputs. Defaults to True.
        tab (str, optional): tab size. Defaults to '    '.
        
    Keyword parameters:
        kws: parameters provided to the `get_script` function, 
            including `sep_step` and `sep_step_end`
    
    TODOs: 
        1. For version control, use https://github.com/jupyterlab/jupyterlab-git.
    """
    packagen=basename(dirname(abspath(".")))
    # packagescriptsp=f"{packagep}/{packagen}"
    packagescriptsp=notebooksdp
    df_outp=f'{packagescriptsp}/.workflow/info.tsv'
    makedirs(df_outp)
    # if notebooksdp is None:
    #     notebooksdp=f'{packagescriptsp}/notebooks'
    # info(packagescriptsp,notebooksdp)
    if scripts:
        # get all the notebooks
        if not ps is None:
            ps=read_ps(ps)
            ps=[abspath(p) for p in ps]
            make_all=False
        else:
            ps=read_ps(f'{notebooksdp}/*ipynb')[::-1]
            make_all=True
        if test: info(len(ps))
        if exists(f'{notebooksdp}/.workflow/config.yaml'):
            from roux.lib.dict import read_dict
            cfg=read_dict(f'{notebooksdp}/.workflow/config.yaml')
            if 'exclude' in cfg:
                ps=[p for p in ps if not basename(p) in cfg['exclude']]
                info(f"remaining few paths after excluding: {len(ps)}")
        df2=get_scripts(ps,force=force,
                        notebook_prefix=notebook_prefix,
                        notebook_suffix=notebook_suffix,
                        sep_step=sep_step,
                        **kws)
        if test: info(df2.shape)
        if not make_all and exists(df_outp):
            df_=pd.read_csv(df_outp,sep='\t').rd.clean()
            df2=df_.loc[~(df_['notebook path'].isin(ps)),:].append(df2).drop_duplicates()
            df2=df2.dropna(subset=['notebook path','rule code'])
            # else:
            #     logging.warning('likely incomplete workflow')
        df2.to_csv(df_outp,sep='\t')
        if test: info(df_outp)
        ## make __init__.py if not exists
        initp=Path(f'{notebooksdp}/lib/__init__.py')
        if not initp.exists(): initp.touch()
    if workflow or (ps is None):
        ## workflow inferred
        if not 'df2' in globals():
            df2=pd.read_csv(df_outp,sep='\t')
        from roux.workflow.io import to_workflow
        to_workflow(df2,workflowp=f'{packagescriptsp}/workflow.py')
    ## make readme
    if todos:
        from roux.workflow.io import to_info
        to_info(p=f'{packagescriptsp}/*{notebook_prefix}*{notebook_suffix}*.ipynb',outp=f'{packagescriptsp}/README.md')
        from roux.lib.text import read_lines
        [print(s) for s in '\n'.join(read_lines(f'{packagescriptsp}/README.md')).split(sep_step) if 'TODO' in s]
    if git:
        from .version import git_commit
        git_commit(packagep,suffix_message='' if not validate else ' (not validated)')
    if clean:
        output = subprocess.check_output(f'grep -Hrn "%run " {packagescriptsp}/lib/*.py',
                                         shell=True,
                                        universal_newlines=True).split('\n')
        print(output)
        


import matplotlib.pyplot as plt
from matplotlib import colors,cm
import numpy as np
import itertools
import seaborn as sns
import pandas as pd

# colors
from matplotlib.colors import to_hex
from matplotlib.colors import ColorConverter 
to_rgb=ColorConverter.to_rgb
def rgbfloat2int(rgb_float):return [int(round(i*255)) for i in rgb_float]
# aliases
rgb2hex=to_hex
hex2rgb=to_rgb   

# colors
def get_colors_default() -> list:
    """get default colors.

    Returns:
        list: colors.
    """
    return plt.rcParams['axes.prop_cycle'].by_key()['color']

def get_ncolors(
    n: int,
    cmap: str='Spectral',
    ceil: bool=False,
    test: bool=False,
    N: int=20,
    out: str='hex',
    **kws_get_cmap_section
    ) -> list:
    """Get colors.

    Args:
        n (int): number of colors to get.
        cmap (str, optional): colormap. Defaults to 'Spectral'.
        ceil (bool, optional): ceil. Defaults to False.
        test (bool, optional): test mode. Defaults to False.
        N (int, optional): number of colors in the colormap. Defaults to 20.
        out (str, optional): output. Defaults to 'hex'.

    Returns:
        list: colors.
    """
    if isinstance(cmap,str):
        cmap = get_cmap_section(cmap, **kws_get_cmap_section)
    elif isinstance(cmap,list):
        cmap=make_cmap(cmap,N=N)
#         cmap = cm.get_cmap(cmap)
    if test:
        print(np.arange(1 if ceil else 0,n+(1 if ceil else 0),1))
        print(np.arange(1 if ceil else 0,n+(1 if ceil else 0),1)/n)
    colors=[cmap(i) for i in np.arange(1 if ceil else 0,n+(1 if ceil else 0),1)/n]
    assert(n==len(colors))
    if out=='hex':
        colors=[rgb2hex(c) for c in colors]
    return colors
              
def get_val2color(
    ds: pd.Series,
    vmin: float=None,
    vmax: float=None,
    cmap: str='Reds'
    ) -> dict:
    """Get color for a value.

    Args:
        ds (pd.Series): values.
        vmin (float, optional): minimum value. Defaults to None.
        vmax (float, optional): maximum value. Defaults to None.
        cmap (str, optional): colormap. Defaults to 'Reds'.

    Returns:
        dict: output.
    """
    if vmin is None:
        vmin=min(ds)
    if vmax is None:
        vmax=max(ds)
    colors = [(plt.get_cmap(cmap) if isinstance(cmap,str) else cmap)((i-vmin)/(vmax-vmin)) for i in ds]
    legend2color = {i:(plt.get_cmap(cmap) if isinstance(cmap,str) else cmap)((i-vmin)/(vmax-vmin)) for i in [vmin,np.mean([vmin,vmax]),vmax]}
    return dict(zip(ds,colors)),legend2color
#    columns=['value','c']             

def saturate_color(color, alpha: float) -> object:
    """Saturate a color.

    Args:
        color (_type_):
        alpha (float): alpha level.

    Returns:
        object: output.

    References: 
        https://stackoverflow.com/a/60562502/3521099
    """
    import colorsys
    from roux.stat.transform import rescale
    alpha=rescale(alpha,[0,2],[1.6,0.4])
    if isinstance(color,str):
        color=hex2rgb(color)
    # convert rgb to hls
    h, l, s = colorsys.rgb_to_hls(*color)
    # manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * alpha), s = s)

def mix_colors(d: dict) -> str:
    """Mix colors.

    Args:
        d (dict): colors to alpha map.

    Returns:
        str: hex color.

    References: 
        https://stackoverflow.com/a/61488997/3521099
    """
    if isinstance(d,list):
        d={k:1.0 for k in d}
    d={k.replace('#',''):d[k] for k in d}
    d_items = sorted(d.items())
    tot_weight = sum(d.values())
    red = int(sum([int(k[:2], 16)*v for k, v in d_items])/tot_weight)
    green = int(sum([int(k[2:4], 16)*v for k, v in d_items])/tot_weight)
    blue = int(sum([int(k[4:6], 16)*v for k, v in d_items])/tot_weight)
    zpad = lambda x: x if len(x)==2 else '0' + x
    c=zpad(hex(red)[2:]) + zpad(hex(green)[2:]) + zpad(hex(blue)[2:])
    return f"#{c}"

# colormaps

def make_cmap(cs: list,N: int=20,**kws):
    """Create a colormap.

    Args:
        cs (list): colors
        N (int, optional): resolution i.e. number of colors. Defaults to 20.

    Returns:
        cmap.
    """
    return colors.LinearSegmentedColormap.from_list("custom", colors=cs, N=N,**kws)

def get_cmap_section(cmap, vmin: float=0.0, vmax: float=1.0, n: int=100) -> object:
    """Get section of a colormap.

    Args:
        cmap (object| str): colormap.
        vmin (float, optional): minimum value. Defaults to 0.0.
        vmax (float, optional): maximum value. Defaults to 1.0.
        n (int, optional): resolution i.e. number of colors. Defaults to 100.

    Returns:
        object: cmap.
    """
    if isinstance(cmap,str):
        cmap=plt.get_cmap(cmap)
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=vmin, b=vmax),
        cmap(np.linspace(vmin, vmax, n)))
    return new_cmap

def append_cmap(
    cmap: str='Reds',
    color: str='#D3DDDC',
    cmap_min: float=0.2,
    cmap_max: float=0.8,
    ncolors: int=100,
    ncolors_min: int=1,
    ncolors_max: int=0,                
    ):
    """Append a color to colormap.

    Args:
        cmap (str, optional): colormap. Defaults to 'Reds'.
        color (str, optional): color. Defaults to '#D3DDDC'.
        cmap_min (float, optional): cmap_min. Defaults to 0.2.
        cmap_max (float, optional): cmap_max. Defaults to 0.8.
        ncolors (int, optional): number of colors. Defaults to 100.
        ncolors_min (int, optional): number of colors minimum. Defaults to 1.
        ncolors_max (int, optional): number of colors maximum. Defaults to 0.

    Returns:
        cmap.

    References: 
        https://matplotlib.org/stable/tutorials/colors/colormap-manipulation.html
    """
    viridis = cm.get_cmap(cmap, ncolors)
    newcolors = viridis(np.linspace(cmap_min, cmap_max, ncolors))
#     pink = np.array([248/256, 24/256, 148/256, 1])
    pink = np.append(np.array(hex2rgb(color))/256,1)
    if ncolors_min!=0:
        newcolors[:ncolors_min, :] = pink
    elif ncolors_max!=0:
        newcolors[ncolors_max:, :] = pink
    newcmp = colors.ListedColormap(newcolors)
    return newcmp


import matplotlib.pyplot as plt
from os.path import splitext

def plot_image(
    imp: str,
    ax: plt.Axes=None,
    force=False,
    margin=0,
    axes=False,
    test=False,
    **kwarg
    ) -> plt.Axes:
    """Plot image e.g. schematic.

    Args:
        imp (str): path of the image.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        force (bool, optional): overwrite output. Defaults to False.
        margin (int, optional): margins. Defaults to 0.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        plt.Axes: `plt.Axes` object.

    :param kwarg: cairosvg: {'dpi':500,'scale':2}; imagemagick: {'trim':False,'alpha':False}

    """
    if splitext(imp)[1]=='.png':
        pngp=imp
    else:
#         if splitext(imp)[1]=='.svg' or force:
#             from roux.lib.figs.convert import svg2png
#             pngp=svg2png(imp,force=force,**kwarg)
#         else:
        from roux.viz.io import to_raster
        pngp=to_raster(imp,force=force,**kwarg)
    ax=plt.subplot() if ax is None else ax
    im=plt.imread(pngp)
    ax.imshow(im,interpolation='catrom')
    ax.set(**{'xticks':[],'yticks':[],'xlabel':'','ylabel':''})
    ax.margins(margin)
    if not axes:
        ax.axis('off')
    return ax 


# from roux.global_imports import *

def read(p):
    """Read a file.

    Args:
        p (str): path.

    Returns:
        s (str): contents.
    """
    with open(p,'r') as f:
        s=f.read()
    return s

from roux.lib.set import read_lines

def get_header(path: str,comment='#',lineno=None):
    """Get the header of a file.

    Args:
        path (str): path.
        comment (str): comment identifier.
        lineno (int): line numbers upto.

    Returns:
        lines (list): header.
    """
    import re
    file = open(path, "r")
    lines=[]
    if not comment is None:
        for i,line in enumerate(file):
            if re.search(f"^{comment}.*", line):
                lines.append(line)
            else:
                break
        if lineno is None:
            return lines
        else:
            return lines[lineno]
    else:
        for i,line in enumerate(file):
            if i==lineno:
                return line
            
## text files
def cat(ps,outp):
    """Concatenate text files.

    Args:
        ps (list): list of paths.
        outp (str): output path.

    Returns:
        outp (str): output path.
    """
    makedirs(outp,exist_ok=True)
    with open(outp, 'w') as outfile:
        for p in ps:
            with open(p) as infile:
                outfile.write(infile.read())    
    return outp
